"""
Understand the verification functions.

The classes
:py:class:`aas_core_codegen.intermediate._types.ImplementationSpecificVerification`
and :py:class:`aas_core_codegen.intermediate._types.PatternVerification` had to be
defined in :py:mod:`aas_core_codegen.intermediate._types` to avoid circular imports.
"""
import collections
import re
from typing import Optional, Tuple, List, MutableMapping, Mapping

from icontract import require, ensure

from aas_core_codegen import parse
from aas_core_codegen.common import Error, Identifier, assert_never
from aas_core_codegen.parse import tree as parse_tree


def _check_support(
    node: parse_tree.Node, argument: Identifier
) -> Optional[List[Error]]:
    """
    Check that we understand the ``node`` in the pattern-matching function.

    The ``argument`` specifies the argument to the verification function, which should
    not be used.
    """
    # NOTE (mristin, 2021-12-19):
    # This run-time check is necessary as we already burned our fingers with it.
    assert isinstance(node, parse_tree.Node), f"{node=}"

    if isinstance(node, parse_tree.Constant):
        if isinstance(node.value, str):
            return None
        else:
            return [
                Error(
                    node.original_node,
                    f"We did not implement the support for non-string constants "
                    f"in pattern matching: {parse_tree.dump(node)}.\n"
                    f"\n"
                    f"Please notify the developers if you need this.",
                )
            ]

    elif isinstance(node, parse_tree.JoinedStr):
        errors = []  # type: List[Error]

        for value in node.values:
            # noinspection PyTypeChecker
            if isinstance(value, str):
                continue
            elif isinstance(value, parse_tree.FormattedValue):
                underlying_errors = _check_support(node=value.value, argument=argument)
                if underlying_errors is not None:
                    errors.extend(underlying_errors)
            else:
                assert_never(value)

        if len(errors) == 0:
            return None

        return errors

    elif isinstance(node, parse_tree.Name):
        if node.identifier == argument:
            return [
                Error(
                    node.original_node,
                    f"The verification arguments, {argument!r}, is not expected "
                    f"to be accessed neither for reading nor for writing.",
                )
            ]
        else:
            return None

    elif isinstance(node, parse_tree.Assignment):
        if not isinstance(node.target, parse_tree.Name):
            return [
                Error(
                    node.target.original_node,
                    f"We currently support only assignments to simple variables, "
                    f"but got: {parse_tree.dump(node.target)}.\n"
                    f"\n"
                    f"Please notify the developers if you need this.",
                )
            ]

        return _check_support(node=node.value, argument=argument)

    else:
        return [
            Error(
                node.original_node,
                f"We did not implement the support for this construct "
                f"in pattern matching: {parse_tree.dump(node)}.\n"
                f"\n"
                f"Please notify the developers if you need this.",
            )
        ]


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _evaluate(
    expr: parse_tree.Expression, state: Mapping[Identifier, str]
) -> Tuple[Optional[str], Optional[Error]]:
    """Evaluate the expression to a string constant."""
    if isinstance(expr, parse_tree.Constant):
        assert isinstance(expr.value, str)
        return expr.value, None

    elif isinstance(expr, parse_tree.Name):
        value = state.get(expr.identifier, None)
        if value is None:
            return (
                None,
                Error(
                    expr.original_node,
                    f"The value of variable {expr.identifier} has not been assigned "
                    f"before",
                ),
            )

        return value, None

    elif isinstance(expr, parse_tree.JoinedStr):
        parts = []  # type: List[str]
        for joined_str_value in expr.values:
            if isinstance(joined_str_value, str):
                parts.append(joined_str_value)

            elif isinstance(joined_str_value, parse_tree.FormattedValue):

                part, error = _evaluate(joined_str_value.value, state=state)
                if error is not None:
                    return None, error

                assert part is not None
                parts.append(part)
            else:
                assert_never(joined_str_value)

        return "".join(parts), None

    else:
        raise AssertionError(f"Unexpected expression: {parse_tree.dump(expr)}")


# fmt: off
@require(
    lambda parsed:
    parsed.verification,
    "Understand only verification functions"
)
@ensure(
    lambda result:
    not (result[2] is None)
    or (
        (result[0] is not None) ^ (result[1] is not None)
    ),
    "Valid match and cause if no error"
)
@ensure(
    lambda result:
    not (result[2] is not None)
    or (result[0] is None and result[1] is None),
    "No match and cause if error"
)
# fmt: on
def try_to_understand(
    parsed: parse.UnderstoodMethod,
) -> Tuple[Optional[str], Optional[Error], Optional[Error]]:
    """
    Try to understand the given verification function as a pattern matching function.

    :param parsed: Verification function as parsed in the parsing phase
    :return: tuple of (pattern, reason for not matching, error)

    We distinguish between two causes why the method could not be understood. The first
    cause, returned as the middle value in the tuple, indicates why the method could
    not be matched, but this non-matching is actually expected.

    The second error, the last value in the return tuple, is an unexpected error. For
    example, if everything looks like a pattern matching function, but there are flags
    in the call to ``match(...)``.
    """
    # Understand only functions that take a single string argument
    if not (
        len(parsed.arguments) == 1
        and isinstance(parsed.arguments[0].type_annotation, parse.AtomicTypeAnnotation)
        and parsed.arguments[0].type_annotation.identifier == "str"
    ):
        return None, Error(parsed.node, "Expected a single ``str`` argument"), None

    # We need to return something,
    if not (
        parsed.returns is not None
        and isinstance(parsed.returns, parse.AtomicTypeAnnotation)
        and parsed.returns.identifier == "bool"
    ):
        return None, Error(parsed.node, "Expected a ``bool`` return value"), None

    if len(parsed.body) == 0:
        return None, Error(parsed.node, "Unexpected empty body"), None

    if not isinstance(parsed.body[-1], parse_tree.Return):
        return (
            None,
            Error(parsed.body[-1].original_node, "Last statement not a a return"),
            None,
        )

    return_node = parsed.body[-1]
    assert isinstance(return_node, parse_tree.Return)
    # noinspection PyUnresolvedReferences
    if return_node.value is None:
        return (
            None,
            Error(parsed.body[-1].original_node, "Expected to return a value"),
            None,
        )

    # noinspection PyUnresolvedReferences
    if return_node.value is None:
        return (
            None,
            Error(parsed.body[-1].original_node, "Expected to return a value"),
            None,
        )

    if (
        isinstance(return_node.value, parse_tree.FunctionCall)
        and return_node.value.name.identifier == "match"
    ):
        return (
            None,
            None,
            Error(
                return_node.original_node,
                "The ``match`` function returns a re.Match object, "
                "but this function expected the return value to be a boolean. "
                "Did you maybe want to write ``return match(...) is not None``?",
            ),
        )

    if not isinstance(return_node.value, parse_tree.IsNotNone):
        return (
            None,
            Error(
                return_node.value.original_node,
                f"Expected to return a ``match(...) is not None``, "
                f"but got: {parse_tree.dump(return_node.value)}",
            ),
            None,
        )

    if not isinstance(return_node.value.value, parse_tree.FunctionCall):
        return (
            None,
            Error(
                return_node.value.value.original_node,
                f"Expected a function call ``match(...)``, "
                f"but got: {parse_tree.dump(return_node.value.value)}",
            ),
            None,
        )

    if return_node.value.value.name.identifier != "match":
        return (
            None,
            Error(
                return_node.value.value.name.original_node,
                f"Expected a call to the function ``match(...)``, "
                f"but got a call "
                f"to function: {return_node.value.value.name.identifier!r}",
            ),
            None,
        )

    # NOTE (mristin, 2021-12-19):
    # From here on we return errors. The verification function looks like a pattern
    # matching so if we can not match (no pun intended), we should signal the user that
    # something was unexpected.

    match_call = return_node.value.value
    assert isinstance(match_call, parse_tree.FunctionCall)
    assert match_call.name.identifier == "match"

    if len(match_call.args) < 2:
        return (
            None,
            None,
            Error(
                match_call.original_node,
                f"The ``match`` function expects two arguments "
                f"(pattern and the text to be matched), "
                f"but you provided {len(match_call.args)} argument(s)",
            ),
        )

    if len(match_call.args) > 2:
        return (
            None,
            None,
            Error(
                match_call.original_node,
                "We do not support calls to the ``match`` function with more than "
                "two arguments (pattern and the text to be matched) "
                "since we could not transpile to other languages and schemas "
                "(*e.g.*, flags such as multi-line matching)",
            ),
        )

    # noinspection PyUnresolvedReferences
    if not (
        isinstance(match_call.args[1], parse_tree.Name)
        and match_call.args[1].identifier == parsed.arguments[0].name
    ):
        return (
            None,
            None,
            Error(
                match_call.original_node,
                f"The second argument to ``match`` function, the text to be matched, "
                f"needs to correspond to the single argument of "
                f"the verification function, {parsed.arguments[0].name!r}. "
                f"Otherwise, we can not transpile the pattern to schemas.\n"
                f"\n"
                f"However, we got: {parse_tree.dump(match_call.args[1])}",
            ),
        )

    # noinspection PyUnresolvedReferences
    if (
        isinstance(match_call.args[0], parse_tree.Name)
        and match_call.args[0].identifier == parsed.arguments[0].name
    ):
        return (
            None,
            None,
            Error(
                match_call.original_node,
                f"The first argument, the pattern, to the ``match`` function "
                f"must not be the argument supplied to "
                f"the verification function, {parsed.arguments[0].name!r}.",
            ),
        )

    # region Check the support of the statements

    errors = []  # type: List[Error]
    for i, stmt in enumerate(parsed.body):
        # Skip the return statement
        if i == len(parsed.body) - 1:
            break

        underlying_errors = _check_support(node=stmt, argument=parsed.arguments[0].name)

        if underlying_errors is not None:
            errors.extend(underlying_errors)

    underlying_errors = _check_support(
        node=match_call.args[0], argument=parsed.arguments[0].name
    )

    if underlying_errors is not None:
        errors.extend(underlying_errors)

    if len(errors) > 0:
        return (
            None,
            None,
            Error(
                parsed.node,
                f"We could not understand "
                f"the pattern matching function {parsed.name!r}",
                errors,
            ),
        )

    # endregion

    # region Re-execute the function to infer the pattern

    state = collections.OrderedDict()  # type: MutableMapping[Identifier, str]

    pattern = None  # type: Optional[str]
    for i, stmt in enumerate(parsed.body):
        if i < len(parsed.body) - 1:
            assert isinstance(stmt, parse_tree.Assignment)
            assert isinstance(stmt.target, parse_tree.Name)

            value, error = _evaluate(expr=stmt.value, state=state)
            if error is not None:
                return None, None, error

            assert value is not None

            state[stmt.target.identifier] = value

        else:
            pattern, error = _evaluate(expr=match_call.args[0], state=state)
            if error is not None:
                return None, None, error

            assert pattern is not None

    # endregion

    assert pattern is not None

    try:
        re.compile(pattern)
    except re.error as exception:
        return (
            None,
            None,
            Error(
                match_call.args[0].original_node,
                f"Failed to compile the pattern with the Python's ``re`` module.\n"
                f"\n"
                f"The evaluated pattern was: {pattern!r}.\n"
                f"The error message was: {exception}",
            ),
        )

    return pattern, None, None



"""Generate the OPC UA Schema node set corresponding to the meta-model."""
import collections
import dataclasses
import enum
import io
import itertools
import re
import xml.dom.minidom
import xml.etree.ElementTree as ET
import xml.sax.saxutils
from typing import (
    TextIO,
    Tuple,
    Optional,
    List,
    MutableMapping,
    Mapping,
    Union,
    OrderedDict,
    Iterable,
    Set,
)

from icontract import ensure, require

import aas_core_codegen.opcua
import aas_core_codegen.opcua.naming as opcua_naming
from aas_core_codegen import (
    run,
    intermediate,
    specific_implementations,
)
from aas_core_codegen.common import (
    Error,
    assert_never,
    Identifier,
    IDENTIFIER_RE,
    Stripped,
)

assert aas_core_codegen.opcua.__doc__ == __doc__

_PRIMITIVE_MAP = {
    intermediate.PrimitiveType.BOOL: "Boolean",
    intermediate.PrimitiveType.INT: "Int64",
    intermediate.PrimitiveType.FLOAT: "Double",
    intermediate.PrimitiveType.STR: "String",
    intermediate.PrimitiveType.BYTEARRAY: "ByteString",
}
assert all(literal in _PRIMITIVE_MAP for literal in intermediate.PrimitiveType)


class _IdentifierMachine:
    """
    Produce stable identifiers for different nodes.

    The Identifier Machine knows nothing about your scheme, so you have to come up
    with your own.

    It will try to be as stable as possible, *i.e.*, the identifiers for the names
    should not change even if the order of :py:func:`obtain` is changed.

    >>> machine = _IdentifierMachine()
    >>> machine.obtain("something")
    207137056

    Obtaining an identifier is idem-potent:

    >>> machine.obtain("something")
    207137056

    Another text gives you another identifier:

    >>> machine.obtain("something_else")
    165180381
    """

    def __init__(self) -> None:
        self._identifier_map = dict()  # type: MutableMapping[str, int]
        self._taken_identifiers = set()  # type: Set[int]

    @staticmethod
    def _hash(text: str) -> int:
        """
        Compute the non-cryptographic hash of the given string.

        We implement our own hash so that the implementation need not change across
        different Python versions, in case they change the hash function.

        >>> _IdentifierMachine._hash("something")
        207137056

        >>> _IdentifierMachine._hash("something_else")
        165180381

        >>> _IdentifierMachine._hash("")
        7
        """
        result = 7
        for character in text:
            result = (result * 31 + ord(character)) % 0x7FFFFFFF

        return result

    def obtain(self, text: str) -> int:
        """
        Assign a slot for the text using hashing.

        In most cases, we do not expect the resulting ID to change.
        """
        identifier = self._identifier_map.get(text, None)
        if identifier is not None:
            return identifier

        identifier = _IdentifierMachine._hash(text)

        while identifier in self._taken_identifiers:
            identifier += 1

        self._taken_identifiers.add(identifier)
        self._identifier_map[text] = identifier

        return identifier


def _generate_aliases() -> ET.Element:
    """Generate the aliases including the primitive values."""
    aliases = ET.Element("Aliases")

    for name, i in (
        ("Boolean", 1),
        ("Int64", 8),
        ("Double", 11),
        ("String", 12),
        ("ByteString", 15),
        ("HasModellingRule", 37),
        ("HasTypeDefinition", 40),
        ("HasSubtype", 45),
        ("HasProperty", 46),
        ("HasComponent", 47),
        ("HasInterface", 17603),
    ):
        alias = ET.Element("Alias", {"Alias": name})
        alias.text = f"i={i}"
        aliases.append(alias)

    return aliases


@dataclasses.dataclass
class _IdentifiersForConstraints:
    """Map the references and object types for constraints to identifiers."""

    constraint_id: int
    constraint_identifier_id: int
    constraint_text_id: int
    has_constraint_id: int


def _generate_for_constraints_and_patterns(
    identifiers_for_constraints: _IdentifiersForConstraints,
) -> List[ET.Element]:
    """Generate the object types and the references to represent constraints."""
    return [
        ET.fromstring(
            f"""\
<UAObjectType
        NodeId="ns=1;i={identifiers_for_constraints.constraint_id}"
        BrowseName="1:Constraint"
>
    <DisplayName>Constraint</DisplayName>
    <References>
        <Reference ReferenceType="HasSubtype" IsForward="false">i=58</Reference>
        <Reference
            ReferenceType="HasProperty"
        >ns=1;i={identifiers_for_constraints.constraint_identifier_id}</Reference>
        <Reference
            ReferenceType="HasProperty"
        >ns=1;i={identifiers_for_constraints.constraint_text_id}</Reference>
    </References>
</UAObjectType>"""
        ),
        ET.fromstring(
            f"""\
<UAVariable
        NodeId="ns=1;i={identifiers_for_constraints.constraint_identifier_id}"
        BrowseName="1:identifier"
        ParentNodeId="ns=1;i={identifiers_for_constraints.constraint_id}"
        DataType="String"
>
    <DisplayName>identifier</DisplayName>
    <References>
        <Reference ReferenceType="HasTypeDefinition">i=68</Reference>
        <Reference ReferenceType="HasModellingRule">i=80</Reference>
    </References>
</UAVariable>"""
        ),
        ET.fromstring(
            f"""\
<UAVariable
        NodeId="ns=1;i={identifiers_for_constraints.constraint_text_id}"
        BrowseName="1:text"
        ParentNodeId="ns=1;i={identifiers_for_constraints.constraint_id}"
        DataType="String"
>
    <DisplayName>text</DisplayName>
    <References>
        <Reference ReferenceType="HasTypeDefinition">i=68</Reference>
        <Reference ReferenceType="HasModellingRule">i=78</Reference>
    </References>
</UAVariable>"""
        ),
        ET.fromstring(
            f"""\
<UAReferenceType
        NodeId="ns=1;i={identifiers_for_constraints.has_constraint_id}"
        BrowseName="1:HasConstraint"
>
    <DisplayName>HasConstraint</DisplayName>
    <References>
        <Reference ReferenceType="HasSubtype" IsForward="false">i=32</Reference>
    </References>
    <InverseName>Constrains</InverseName>
</UAReferenceType>"""
        ),
    ]


#: All the classes whose instances have an ``i=...`` in the node set.
_OpcUaIdentifiable = Union[
    intermediate.Enumeration,
    intermediate.ConstrainedPrimitive,
    intermediate.Class,
    intermediate.Interface,
    intermediate.Property,
    intermediate.Invariant,
]


def _reference(
    reference_type: str, target: str, is_forward: Optional[bool] = None
) -> ET.Element:
    """Create a ``<Reference>`` element."""
    attrib = collections.OrderedDict([("ReferenceType", reference_type)])
    if is_forward is not None:
        attrib["IsForward"] = "true" if is_forward else "false"

    reference_el = ET.Element("Reference", attrib)
    reference_el.text = target

    return reference_el


def _localized_text(text: str) -> ET.Element:
    """Create a nested ``<uax:LocalizedText>``."""
    localized_text_el = ET.Element("uax:LocalizedText")

    text_el = ET.Element("uax:Text")
    localized_text_el.append(text_el)

    text_el.text = text

    return localized_text_el


def _generate_for_enum(
    enumeration: intermediate.Enumeration,
    identifier_map: Mapping[_OpcUaIdentifiable, int],
    identifier_machine: _IdentifierMachine,
    name_prefix: Identifier,
) -> List[ET.Element]:
    """Define the enumeration as OPC UA EnumStrings."""
    enum_name = opcua_naming.enum_name(enumeration.name, name_prefix)

    result = []  # type: List[ET.Element]

    comment_starts = ET.Comment(f"{enum_name} starts.")
    comment_starts.tail = "\n"
    result.append(comment_starts)

    enum_id = identifier_map[enumeration]

    def generate_data_type() -> ET.Element:
        """Generate the element defining the data type."""
        data_type_el = ET.Element(
            "UADataType",
            collections.OrderedDict(
                [("NodeId", f"ns=1;i={enum_id}"), ("BrowseName", f"1:{enum_name}")]
            ),
        )

        display_name_el = ET.Element("DisplayName")
        data_type_el.append(display_name_el)
        display_name_el.text = enum_name

        references_el = ET.Element("References")
        data_type_el.append(references_el)
        references_el.append(
            _reference(reference_type="HasSubtype", target="i=29", is_forward=False)
        )

        definition_el = ET.Element("Definition", {"Name": f"1:{enum_name}"})
        data_type_el.append(definition_el)

        for i, literal in enumerate(enumeration.literals):
            field = ET.Element(
                "Field",
                collections.OrderedDict(
                    [
                        ("Name", opcua_naming.enum_literal_name(literal.name)),
                        ("Value", str(i)),
                    ]
                ),
            )
            definition_el.append(field)

        return data_type_el

    result.append(generate_data_type())

    def generate_enum_strings() -> ET.Element:
        """Generate the EnumString for the enumeration."""
        variable_id = identifier_machine.obtain(f"{enumeration.name}:variable")

        variable_el = ET.Element(
            "UAVariable",
            collections.OrderedDict(
                [
                    ("DataType", "LocalizedText"),
                    ("ValueRank", "1"),
                    ("NodeId", f"ns=1;i={variable_id}"),
                    ("ArrayDimensions", str(len(enumeration.literals))),
                    ("BrowseName", "EnumStrings"),
                    ("ParentNodeId", f"ns=1;i={enum_id}"),
                ]
            ),
        )

        display_name_el = ET.Element("DisplayName")
        variable_el.append(display_name_el)
        display_name_el.text = "EnumStrings"

        references_el = ET.Element("References")
        variable_el.append(references_el)
        references_el.append(
            _reference("HasProperty", target=f"ns=1;i={enum_id}", is_forward=False)
        )
        references_el.append(_reference("HasTypeDefinition", target="i=68"))
        references_el.append(_reference("HasModellingRule", target="i=78"))

        value_el = ET.Element("Value")
        variable_el.append(value_el)

        list_of_localized_text_el = ET.Element("uax:ListOfLocalizedText")
        value_el.append(list_of_localized_text_el)

        for literal in enumeration.literals:
            list_of_localized_text_el.append(_localized_text(literal.value))

        return variable_el

    result.append(generate_enum_strings())

    comment_ends = ET.Comment(f"{enum_name} ends.")
    comment_ends.tail = "\n"
    result.append(comment_ends)

    return result


_CONSTRAINT_ID_PREFIX_RE = re.compile(r"Constraint \s*(?P<identifier>[^:]+)\s*:")


def _extract_constraint_identifier(description: str) -> Optional[str]:
    """
    Try to extract the constraint identifier from the invariant description.

    >>> _extract_constraint_identifier("Constraint 1: bla bla bla")
    '1'

    >>> _extract_constraint_identifier("Constraint  1 : bla bla bla")
    '1'

    >>> _extract_constraint_identifier("Bla bla bla")

    >>> _extract_constraint_identifier("Constraint 2: Name with at most 9 characters.")
    '2'
    """
    match = _CONSTRAINT_ID_PREFIX_RE.match(description)

    if match is None:
        return None

    return match.group("identifier").strip()


def _value_string(text: str) -> ET.Element:
    """Generate an ``<Value>`` element with the embedded string element."""
    value_el = ET.Element("Value")

    uax_string_el = ET.Element("uax:String")
    value_el.append(uax_string_el)
    uax_string_el.text = text

    return value_el


def _generate_for_invariant(
    invariant: intermediate.Invariant,
    identifier_map: Mapping[_OpcUaIdentifiable, int],
    identifiers_for_constraints: _IdentifiersForConstraints,
    identifier_machine: _IdentifierMachine,
) -> List[ET.Element]:
    """Generate the constraint object for the invariant."""
    assert (
        invariant in identifier_map
    ), f"Invariant {invariant.description!r} missing in identifier map"

    invariant_id = identifier_map[invariant]

    constraint_identifier = _extract_constraint_identifier(invariant.description)

    if constraint_identifier is None:
        browse_name = f"ConstraintUnlabeled{invariant_id}"
    else:
        browse_name = opcua_naming.constraint_browser_name(
            Stripped(constraint_identifier)
        )

    object_el = ET.fromstring(
        f"""\
<UAObject
        NodeId="ns=1;i={invariant_id}"
        BrowseName="1:{browse_name}"
        ParentNodeId="i=85"
>
    <DisplayName>{browse_name}</DisplayName>
    <References>
        <Reference ReferenceType="Organizes" IsForward="false">i=85</Reference>
        <Reference
                ReferenceType="HasTypeDefinition"
        >ns=1;i={identifiers_for_constraints.constraint_id}</Reference>
    </References>
</UAObject>"""
    )

    text_id = identifier_machine.obtain(f"{invariant_id}:text")

    text_variable_el = ET.fromstring(
        f"""\
<UAVariable
        DataType="String"
        NodeId="ns=1;i={text_id}"
        BrowseName="1:text"
        ParentNodeId="ns=1;i={invariant_id}"
>
    <DisplayName>text</DisplayName>
    <References>
        <Reference
                ReferenceType="HasProperty"
                IsForward="false"
        >ns=1;i={invariant_id}</Reference>
        <Reference ReferenceType="HasTypeDefinition">i=63</Reference>
    </References>
</UAVariable>"""
    )

    # NOTE (mristin):
    # We have to create the ``<Value>`` ourselves as we use a namespace alias and
    # the ElementTree can not parse the namespace aliases from string.
    text_variable_el.append(_value_string(invariant.description))

    result = [object_el, text_variable_el]

    if constraint_identifier is not None:
        identifier_id = identifier_machine.obtain(f"{invariant_id}:identifier")
        identifier_variable_el = ET.fromstring(
            f"""\
<UAVariable
        DataType="String"
        NodeId="ns=1;i={identifier_id}"
        BrowseName="1:identifier"
        ParentNodeId="ns=1;i={invariant_id}"
>
    <DisplayName>identifier</DisplayName>
    <References>
        <Reference
                ReferenceType="HasProperty"
                IsForward="false"
        >ns=1;i={invariant_id}</Reference>
        <Reference ReferenceType="HasTypeDefinition">i=63</Reference>
    </References>
</UAVariable>"""
        )

        # NOTE (mristin):
        # We have to create the ``<Value>`` ourselves as we use a namespace alias and
        # the ElementTree can not parse the namespace aliases from string.
        identifier_variable_el.append(_value_string(constraint_identifier))

        result.append(identifier_variable_el)

    return result


def _generate_for_constrained_primitive(
    constrained_primitive: intermediate.ConstrainedPrimitive,
    identifier_map: Mapping[_OpcUaIdentifiable, int],
    identifiers_for_constraints: _IdentifiersForConstraints,
    name_prefix: Identifier,
) -> ET.Element:
    """Define the constrained primitive and link it with constraints."""
    constrained_primitive_name = opcua_naming.constrained_primitive_name(
        constrained_primitive.name, prefix=name_prefix
    )
    constrained_primitive_id = identifier_map[constrained_primitive]

    root = ET.fromstring(
        f"""\
<UADataType
        NodeId="ns=1;i={constrained_primitive_id}"
        BrowseName="1:{constrained_primitive_name}"
>
    <DisplayName>{constrained_primitive_name}</DisplayName>
    <References>
        <Reference
                ReferenceType="HasSubtype"
                IsForward="false"
        >{_PRIMITIVE_MAP[constrained_primitive.constrainee]}</Reference>
    </References>
</UADataType>"""
    )

    references_el = root.find("References")
    assert references_el is not None, "Expected <References> in the node"

    for invariant in constrained_primitive.invariants:
        # NOTE (mristin):
        # We do not model inheritance between the constrained primitives as this
        # is not possible in OPC UA. Hence, we define all the constraints for each
        # constrained primitive. This causes a lot of repetition, but we found no other
        # way around it.

        invariant_id = identifier_map[invariant]

        references_el.append(
            ET.fromstring(
                f"""\
<Reference
        ReferenceType="ns=1;i={identifiers_for_constraints.has_constraint_id}"
>ns=1;i={invariant_id}</Reference>"""
            )
        )

    return root


class _PropertyReferenceType(enum.Enum):
    PROPERTY = 0
    COMPONENT = 1


def _determine_property_reference_type(
    prop: intermediate.Property,
) -> _PropertyReferenceType:
    """
    Determine how to map the given property.

    In OPC UA, we distinguish between properties — attributes with simple data types or
    arrays of simple data types — and components — references to instances
    or aggregations of instances.
    """
    type_anno = intermediate.beneath_optional(prop.type_annotation)

    if (intermediate.try_primitive_type(type_anno) is not None) or (
        isinstance(type_anno, intermediate.ListTypeAnnotation)
        and intermediate.try_primitive_type(type_anno.items) is not None
    ):
        return _PropertyReferenceType.PROPERTY

    return _PropertyReferenceType.COMPONENT


def _try_primitive_type(
    type_annotation: intermediate.TypeAnnotationUnion,
) -> Optional[intermediate.PrimitiveType]:
    """
    Inspect the type annotation and determine the underlying primitive type, if any.

    The primitive type can either be in the annotation itself, beneath an optional
    or beneath a list.

    This is different to :py:func:`intermediate.try_primitive_typ` since this function
    considers lists as well. This is because OPC UA does not distinguish between
    scalars and arrays when it comes to primitive types.
    """
    type_anno = intermediate.beneath_optional(type_annotation)

    # NOTE (mristin):
    # We make sure that we do not use the variable unintentionally.
    del type_annotation

    if isinstance(type_anno, intermediate.PrimitiveTypeAnnotation):
        return type_anno.a_type

    elif isinstance(type_anno, intermediate.OurTypeAnnotation) and isinstance(
        type_anno.our_type, intermediate.ConstrainedPrimitive
    ):
        return type_anno.our_type.constrainee

    elif isinstance(type_anno, intermediate.ListTypeAnnotation) and isinstance(
        type_anno.items, intermediate.PrimitiveTypeAnnotation
    ):
        return type_anno.items.a_type

    else:
        return None


def _generate_references_for_properties(
    properties: Iterable[intermediate.Property],
    identifier_map: Mapping[_OpcUaIdentifiable, int],
) -> List[ET.Element]:
    """
    Generate the ``<Reference>`` entries for the given properties.

    ``prop_to_id`` maps each property to an OPC UA identifier.
    """
    result = []  # type: List[ET.Element]

    for prop in properties:
        prop_id = identifier_map[prop]

        property_reference_type = _determine_property_reference_type(prop)
        if property_reference_type is _PropertyReferenceType.PROPERTY:
            result.append(_reference("HasProperty", f"ns=1;i={prop_id}"))
        elif property_reference_type is _PropertyReferenceType.COMPONENT:
            result.append(_reference("HasComponent", f"ns=1;i={prop_id}"))
        else:
            assert_never(property_reference_type)

    return result


@ensure(lambda result: result.tag in ("UAVariable", "UAObject"))
def _generate_for_property(
    prop: intermediate.Property,
    prop_id: int,
    parent_id: int,
    identifier_map: Mapping[_OpcUaIdentifiable, int],
) -> ET.Element:
    """Generate the variable element corresponding to the property."""
    prop_name = opcua_naming.property_name(prop.name)

    root = ET.Element(
        "",
        attrib=collections.OrderedDict(
            [
                ("NodeId", f"ns=1;i={prop_id}"),
                ("BrowseName", f"1:{prop_name}"),
                ("ParentNodeId", f"ns=1;i={parent_id}"),
            ]
        ),
    )

    display_name_el = ET.Element("DisplayName")
    root.append(display_name_el)
    display_name_el.text = prop_name

    if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
        type_anno = prop.type_annotation.value
    else:
        type_anno = prop.type_annotation

    if isinstance(type_anno, intermediate.OptionalTypeAnnotation):
        raise AssertionError(
            "NOTE (mristin): We do not handle optional optionals, but you "
            f"specified type annotation {prop.type_annotation} for the property "
            f"{prop.name!r}."
        )

    if isinstance(type_anno, intermediate.ListTypeAnnotation):
        type_anno = type_anno.items

    elif isinstance(
        type_anno,
        (intermediate.PrimitiveTypeAnnotation, intermediate.OurTypeAnnotation),
    ):
        pass

    else:
        assert_never(prop.type_annotation)

    if not isinstance(
        type_anno,
        (intermediate.PrimitiveTypeAnnotation, intermediate.OurTypeAnnotation),
    ):
        raise AssertionError(
            f"NOTE (mristin): We only implemented optional and mandatory lists of "
            f"primitives and lists of model instances, but you have "
            f"a property {prop.name} with type: {prop.type_annotation}. "
            f"Please contact the developers if you need this feature."
        )

    primitive_type = _try_primitive_type(type_anno)

    is_list = isinstance(
        intermediate.beneath_optional(prop.type_annotation),
        intermediate.ListTypeAnnotation,
    )
    is_optional = isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation)

    references_el = ET.Element("References")
    if primitive_type is not None:
        if is_optional:
            # NOTE (mristin):
            # This corresponds to ``Optional``.
            references_el.append(_reference("HasModellingRule", target="i=80"))
        else:
            # NOTE (mristin):
            # This corresponds to ``Mandatory``.
            references_el.append(_reference("HasModellingRule", target="i=78"))
    else:
        if is_list:
            if is_optional:
                # NOTE (mristin):
                # This corresponds to ``OptionalPlaceholder``.
                references_el.append(_reference("HasModellingRule", target="i=11508"))
            else:
                # NOTE (mristin):
                # This corresponds to ``MandatoryPlaceholder``.
                references_el.append(_reference("HasModellingRule", target="i=11510"))
        else:
            if is_optional:
                # NOTE (mristin):
                # This corresponds to ``Optional``.
                references_el.append(_reference("HasModellingRule", target="i=80"))
            else:
                # NOTE (mristin):
                # This corresponds to ``Mandatory``.
                references_el.append(_reference("HasModellingRule", target="i=78"))

    if primitive_type is not None:
        if isinstance(type_anno, intermediate.PrimitiveTypeAnnotation):
            root.attrib["DataType"] = _PRIMITIVE_MAP[primitive_type]

        elif isinstance(type_anno, intermediate.OurTypeAnnotation) and isinstance(
            type_anno.our_type, intermediate.ConstrainedPrimitive
        ):
            root.attrib["DataType"] = f"ns=1;i={identifier_map[type_anno.our_type]}"

        else:
            raise AssertionError(
                f"Unexpected primitive type in the property {prop.name!r} with "
                f"type annotation {prop.type_annotation} where the type annotation "
                f"beneath any Optional or List was: {type_anno}"
            )

        references_el.append(_reference("HasTypeDefinition", target="i=68"))

        if is_list:
            root.attrib["ValueRank"] = "1"

        root.tag = "UAVariable"
    else:
        if isinstance(type_anno, intermediate.PrimitiveTypeAnnotation):
            raise AssertionError(
                "The case of the primitive type should have been handled before."
            )
        elif isinstance(type_anno, intermediate.OurTypeAnnotation):
            if isinstance(type_anno.our_type, intermediate.Enumeration):
                root.tag = "UAVariable"

                enum_id = identifier_map[type_anno.our_type]

                root.attrib["DataType"] = f"ns=1;i={enum_id}"

                references_el.append(_reference("HasTypeDefinition", target="i=62"))

            elif isinstance(type_anno.our_type, intermediate.ConstrainedPrimitive):
                raise AssertionError(
                    "The case of the primitive type should have been handled before."
                )

            elif isinstance(
                type_anno.our_type,
                (intermediate.AbstractClass, intermediate.ConcreteClass),
            ):
                root.tag = "UAObject"

                if type_anno.our_type.interface is not None:
                    # NOTE (mristin):
                    # We specify the type as ``BaseObjectType`` since multiple
                    # inheritance is not possible.
                    references_el.append(_reference("HasTypeDefinition", target="i=58"))

                    interface_id = identifier_map[type_anno.our_type.interface]
                    references_el.append(
                        _reference("HasInterface", target=f"ns=1;i={interface_id}")
                    )

                    for ancestor_cls in type_anno.our_type.ancestors:
                        assert ancestor_cls.interface is not None

                        ancestor_id = identifier_map[ancestor_cls.interface]
                        references_el.append(
                            _reference("HasInterface", target=f"ns=1;i={ancestor_id}")
                        )

                    references_el.extend(
                        _generate_references_for_properties(
                            # NOTE (mristin):
                            # We reference *all* the properties of the object since
                            # the type definition is set to ``BaseObjectType``, but
                            # needs to satisfy all the interfaces.
                            properties=type_anno.our_type.properties,
                            identifier_map=identifier_map,
                        )
                    )

                else:
                    references_el.append(
                        _reference(
                            "HasTypeDefinition",
                            target=f"ns=1;i={identifier_map[type_anno.our_type]}",
                        )
                    )
            else:
                assert_never(type_anno.our_type)
        else:
            assert_never(type_anno)

    root.append(references_el)

    return root


def _generate_definitions_for_properties(
    parent_id: int,
    properties: Iterable[intermediate.Property],
    identifier_map: Mapping[_OpcUaIdentifiable, int],
) -> List[ET.Element]:
    """
    Generate the definition elements for all the properties.

    ``prop_to_id`` maps the properties to OPC UA identifiers.
    """
    result = []  # type: List[ET.Element]

    for prop in properties:
        prop_id = identifier_map[prop]

        property_el = _generate_for_property(
            prop=prop,
            prop_id=prop_id,
            parent_id=parent_id,
            identifier_map=identifier_map,
        )

        result.append(property_el)

    return result


def _generate_for_interface(
    interface: intermediate.Interface,
    identifier_map: Mapping[_OpcUaIdentifiable, int],
    identifiers_for_constraints: _IdentifiersForConstraints,
    name_prefix: Identifier,
) -> List[ET.Element]:
    """Generate the definition for the given interface."""
    interface_name = opcua_naming.interface_name(interface.name, name_prefix)

    comment_starts = ET.Comment(f"{interface_name} starts.")
    comment_starts.tail = "\n"

    result = [comment_starts]  # type: List[ET.Element]

    interface_id = identifier_map[interface]

    object_type_el = ET.Element(
        "UAObjectType",
        collections.OrderedDict(
            [
                ("NodeId", f"ns=1;i={interface_id}"),
                ("BrowseName", f"1:{interface_name}"),
                ("IsAbstract", "true"),
            ]
        ),
    )
    result.append(object_type_el)

    display_name_el = ET.Element("DisplayName")
    display_name_el.text = interface_name
    object_type_el.append(display_name_el)

    strictly_interface_properties = [
        prop for prop in interface.properties if prop.specified_for is interface.base
    ]

    references_el = ET.Element("References")
    object_type_el.append(references_el)

    references_el.append(_reference("HasSubtype", target="i=17602", is_forward=False))

    for invariant in interface.base.invariants:
        if invariant.specified_for is not interface.base:
            continue

        invariant_id = identifier_map[invariant]

        references_el.append(
            _reference(
                reference_type=f"ns=1;i={identifiers_for_constraints.has_constraint_id}",
                target=f"ns=1;i={invariant_id}",
            )
        )

    references_el.extend(
        _generate_references_for_properties(
            properties=strictly_interface_properties, identifier_map=identifier_map
        )
    )

    result.extend(
        _generate_definitions_for_properties(
            parent_id=interface_id,
            properties=strictly_interface_properties,
            identifier_map=identifier_map,
        )
    )

    comment_ends = ET.Comment(f"{interface_name} ends.")
    comment_ends.tail = "\n"
    result.append(comment_ends)

    return result


def _generate_for_concrete_class(
    cls: intermediate.ConcreteClass,
    identifier_map: Mapping[_OpcUaIdentifiable, int],
    identifiers_for_constraints: _IdentifiersForConstraints,
    name_prefix: Identifier,
) -> List[ET.Element]:
    """Generate the definition for the given concrete class."""
    cls_name = opcua_naming.class_name(cls.name, name_prefix)
    cls_id = identifier_map[cls]

    result = [ET.Comment(f"{cls_name} starts.")]  # type: List[ET.Element]

    object_type_el = ET.Element(
        "UAObjectType",
        attrib=collections.OrderedDict(
            [("NodeId", f"ns=1;i={cls_id}"), ("BrowseName", f"1:{cls_name}")]
        ),
    )

    display_name_el = ET.Element("DisplayName")
    display_name_el.text = cls_name
    object_type_el.append(display_name_el)

    references_el = ET.Element("References")
    references_el.append(_reference("HasSubtype", target="i=58", is_forward=False))

    if cls.interface is not None:
        interface_id = identifier_map[cls.interface]
        references_el.append(_reference("HasInterface", f"ns=1;i={interface_id}"))

    for ancestor_cls in cls.ancestors:
        assert ancestor_cls.interface is not None

        assert ancestor_cls.interface in identifier_map, (
            f"The OPC UA identifier is missing for the interface corresponding to "
            f"the ancestor class {ancestor_cls.name!r} of the class {cls.name!r}"
        )

        ancestor_id = identifier_map[ancestor_cls.interface]
        references_el.append(_reference("HasInterface", f"ns=1;i={ancestor_id}"))

    if cls.interface is None:
        # NOTE (mristin):
        # We have to reference the constraints here since this concrete class has no
        # interface, so no constraints have been referenced thus far. That is, we define
        # the constraints in interfaces, but this class has none, so we have to
        # reference them in the ``ObjectType`` corresponding to the class.
        for invariant in cls.invariants:
            if invariant.specified_for is not cls:
                continue

            invariant_id = identifier_map[invariant]

            references_el.append(
                _reference(
                    reference_type=f"ns=1;i={identifiers_for_constraints.has_constraint_id}",
                    target=f"ns=1;i={invariant_id}",
                )
            )

    else:
        # NOTE (mristin):
        # We assume that the object type needs to implement all the constraints imposed
        # on its interfaces. Since we define an interface for any concrete class with
        # descendants, we do not have to reference the constraints here.
        pass

    strictly_cls_properties = [
        prop for prop in cls.properties if prop.specified_for is cls
    ]

    references_el.extend(
        _generate_references_for_properties(
            properties=strictly_cls_properties, identifier_map=identifier_map
        )
    )

    object_type_el.append(references_el)
    result.append(object_type_el)

    # NOTE (mristin):
    # We define the properties only if they have not been already defined
    # for an interface.
    # Otherwise, we refer to the properties of the interfaces through
    # ``HasComponent``/``HasProperty`` reference.
    if cls.interface is None:
        result.extend(
            _generate_definitions_for_properties(
                parent_id=cls_id,
                properties=strictly_cls_properties,
                identifier_map=identifier_map,
            )
        )

    result.append(ET.Comment(f"{cls_name} ends."))

    return result


class _NamespaceDeclarations:
    #: URL of the main namespace
    main: str

    #: URL to namespace alias
    url_to_alias: Mapping[str, str]

    #: Namespace alias to URL
    alias_to_url: OrderedDict[str, str]

    # fmt: off
    @require(
        lambda url_to_alias, alias_to_url:
        all(
            alias_to_url[alias] == url
            for url, alias in url_to_alias.items()
        )
    )
    @require(
        lambda url_to_alias, alias_to_url:
        all(
            url_to_alias[url] == alias
            for alias, url in alias_to_url.items()
        )
    )
    # fmt: on
    def __init__(
        self,
        main: str,
        url_to_alias: Mapping[str, str],
        alias_to_url: OrderedDict[str, str],
    ) -> None:
        """Initialize with the given values."""
        self.main = main
        self.url_to_alias = url_to_alias
        self.alias_to_url = alias_to_url


def _extractNamespaceDeclarationsFromXML(
    text: str,
) -> Tuple[Optional[_NamespaceDeclarations], Optional[str]]:
    """
    Extract the namespace declarations from the given XML document.

    Return the parsed declarations, or error, if any.
    """
    minidom_doc = xml.dom.minidom.parseString(text)

    main = None  # type: Optional[str]
    url_to_alias = dict()  # type: MutableMapping[str, str]
    alias_to_url = collections.OrderedDict()  # type: OrderedDict[str, str]

    for attribute, value in minidom_doc.documentElement.attributes.items():
        if attribute == "xmlns":
            main = value
        elif attribute.startswith("xmlns:"):
            alias = attribute[len("xmlns:") :]
            url_to_alias[value] = alias
            alias_to_url[alias] = value
        else:
            # NOTE (mristin):
            # This attribute is otherwise irrelevant.
            pass

    if main is None:
        return None, "The main namespace is missing"

    return (
        _NamespaceDeclarations(
            main=main, url_to_alias=url_to_alias, alias_to_url=alias_to_url
        ),
        None,
    )


INDENT = "  "

# NOTE (mristin):
# The ElementTree library is very peculiar when it comes to namespaces. The namespace
# URLs are directly inserted as prefixes to tag names, even when the user specifies
# an alias.
_ET_NAMESPACE_PREFIX_RE = re.compile(r"^\{(?P<namespace>[^}]*)}")


def _render(
    element: ET.Element,
    writer: TextIO,
    namespace_declarations: _NamespaceDeclarations,
    level: int = 0,
) -> None:
    """
    Render the given XML tree starting with the ``element`` at the root.

    We designed the rendering such that the XMLs are easy to read and diff.
    """
    indention = INDENT * level

    if element.tag is ET.Comment:  # type: ignore
        writer.write(f"{indention}<!-- {element.text} -->\n")
    else:
        if len(element) > 0:
            if element.text is not None and not element.text.isspace():
                raise ValueError(
                    f"Unexpected element with children "
                    f"and non-whitespace text: {element}; "
                    f"the text was: {element.text!r}"
                )

            if element.tail is not None and not element.tail.isspace():
                raise ValueError(
                    f"Unexpected element with children "
                    f"and non-whitespace tail: {element}; "
                    f"the tail was: {element.tail!r}"
                )

        ns_match = _ET_NAMESPACE_PREFIX_RE.match(element.tag)
        if ns_match is not None:
            ns_url = ns_match.group("namespace")

            if ns_url == namespace_declarations.main:
                # NOTE (mristin):
                # We add 2 for the enclosing ``{`` and ``}``.
                name = element.tag[len(ns_url) + 2 :]
            else:
                ns_alias = namespace_declarations.url_to_alias.get(ns_url, None)
                if ns_alias is None:
                    raise ValueError(
                        f"Unexpected namespace URL in the ET element {element}. "
                        f"The declared namespaces aliases "
                        f"were: {namespace_declarations.url_to_alias}"
                    )

                # NOTE (mristin):
                # We add 2 for the enclosing ``{`` and ``}``.
                local_name = element.tag[len(ns_url) + 2 :]
                name = f"{ns_alias}:{local_name}"
        else:
            name = element.tag

        text = None  # type: Optional[str]
        if element.text is not None and not element.text.isspace():
            text = element.text

        if text is None and len(element) == 0:
            if len(element.attrib) == 0:
                writer.write(f"{indention}<{name} />\n")
            else:
                writer.write(f"{indention}<{name}\n")
                for attrib, value in element.attrib.items():
                    quoted_value = xml.sax.saxutils.quoteattr(value)
                    writer.write(f"{indention}{INDENT}{attrib}={quoted_value}\n")

                writer.write(f"{indention}/>\n")

        elif text is not None and len(element) == 0:
            escaped_text = xml.sax.saxutils.escape(text)

            if len(element.attrib) == 0:
                writer.write(f"{indention}<{name}>{escaped_text}</{name}>\n")
            else:
                writer.write(f"{indention}<{name}\n")

                for attrib, value in element.attrib.items():
                    quoted_value = xml.sax.saxutils.quoteattr(value)
                    writer.write(f"{indention}{INDENT}{attrib}={quoted_value}\n")

                writer.write(f"{indention}>{escaped_text}</{name}>\n")

        elif text is None and len(element) > 0:
            if len(element.attrib) == 0:
                writer.write(f"{indention}<{name}>\n")

                for child in element:
                    _render(
                        element=child,
                        writer=writer,
                        namespace_declarations=namespace_declarations,
                        level=level + 1,
                    )

                writer.write(f"{indention}</{name}>\n")
            else:
                writer.write(f"{indention}<{name}\n")

                for attrib, value in element.attrib.items():
                    quoted_value = xml.sax.saxutils.quoteattr(value)
                    writer.write(f"{indention}{INDENT}{attrib}={quoted_value}\n")

                writer.write(f"{indention}>\n")

                for child in element:
                    _render(
                        element=child,
                        writer=writer,
                        namespace_declarations=namespace_declarations,
                        level=level + 1,
                    )

                writer.write(f"{indention}</{name}>\n")

        else:
            raise AssertionError("Unexpected execution path")


@ensure(lambda result: not (result[1] is not None) or (len(result[1]) >= 1))
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
@ensure(lambda result: not (result[0] is not None) or (result[0].endswith("\n")))
def _generate(
    symbol_table: intermediate.SymbolTable,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """Generate tne node set according to the symbol table."""
    base_nodeset_key = specific_implementations.ImplementationKey("base_nodeset.xml")

    base_nodeset_text = spec_impls.get(base_nodeset_key, None)
    if base_nodeset_text is None:
        return None, [
            Error(
                None,
                f"The implementation snippet for the base OPC UA nodeset "
                f"is missing: {base_nodeset_key}",
            )
        ]

    name_prefix_key = specific_implementations.ImplementationKey("name_prefix.txt")

    name_prefix_text = spec_impls.get(name_prefix_key, None)
    if name_prefix_text is None:
        return None, [
            Error(
                None,
                f"The implementation snippet for the type name prefix "
                f"is missing: {name_prefix_key}",
            )
        ]

    if not IDENTIFIER_RE.match(name_prefix_text):
        return None, [
            Error(
                None,
                f"The implementation snippet for the type name prefix "
                f"at {name_prefix_key} is invalid: {name_prefix_text}",
            )
        ]

    name_prefix = Identifier(name_prefix_text)

    namespace_declarations, error = _extractNamespaceDeclarationsFromXML(
        base_nodeset_text
    )
    if error is not None:
        return None, [
            Error(
                None,
                f"The namespaces could not be extracted from the implementation snippet "
                f"for the base OPC UA nodeset {base_nodeset_key}: {error}",
            )
        ]

    assert namespace_declarations is not None

    if "uax" not in namespace_declarations.alias_to_url:
        return None, [
            Error(
                None,
                f"The namespace alias for 'uax' is missing in "
                f"the implementation snippet for the base OPC UA "
                f"nodeset: {base_nodeset_key}",
            )
        ]

    try:
        root = ET.fromstring(base_nodeset_text)
    except Exception as err:
        return None, [
            Error(
                None,
                f"Failed to parse the base nodeset XML out of "
                f"the snippet {base_nodeset_key}: {err}",
            )
        ]

    identifier_machine = _IdentifierMachine()

    identifier_map = (
        collections.OrderedDict()
    )  # type: MutableMapping[_OpcUaIdentifiable, int]

    root.append(_generate_aliases())

    # region Generate identifiers

    identifiers_for_constraints = _IdentifiersForConstraints(
        # NOTE (mristin):
        # We prefix here with ``_`` in order to avoid conflicts with AAS class names.
        constraint_id=identifier_machine.obtain("_Constraint"),
        constraint_identifier_id=identifier_machine.obtain("_Constraint.identifier"),
        constraint_text_id=identifier_machine.obtain("_Constraint.text"),
        has_constraint_id=identifier_machine.obtain("_HasConstraint"),
    )

    # noinspection SpellCheckingInspection
    invariants = [
        invariant
        for invariantable in itertools.chain(
            symbol_table.classes, symbol_table.constrained_primitives
        )
        for invariant in invariantable.invariants
        if invariant.specified_for is invariantable
    ]

    observed_invariant_uid_set = set()  # type: Set[str]

    for invariant in invariants:
        # NOTE (mristin):
        # We rely on
        # :py:func:`intermediate._translate._verify_invariant_descriptions_unique`
        # that the invariant descriptions are unique.

        invariant_uid = (
            f"{invariant.specified_for.name}" f":invariant:{invariant.description}"
        )

        assert (
            invariant_uid not in observed_invariant_uid_set
        ), f"Unexpected duplicate invariant ID for: {invariant_uid}"
        observed_invariant_uid_set.add(invariant_uid)

        identifier_map[invariant] = identifier_machine.obtain(invariant_uid)

    for enumeration in symbol_table.enumerations:
        identifier_map[enumeration] = identifier_machine.obtain(enumeration.name)

    for constrained_primitive in symbol_table.constrained_primitives:
        identifier_map[constrained_primitive] = identifier_machine.obtain(
            constrained_primitive.name
        )

    for cls in symbol_table.classes:
        if cls.interface is not None:
            identifier_map[cls.interface] = identifier_machine.obtain(
                # NOTE (mristin):
                # We prefix with ``interface:`` to avoid conflicts with
                # concrete classes.
                f"interface:{cls.name}"
            )

    for concrete_cls in symbol_table.concrete_classes:
        identifier_map[concrete_cls] = identifier_machine.obtain(concrete_cls.name)

    for cls in symbol_table.classes:
        for prop in cls.properties:
            if prop.specified_for is not cls:
                continue

            identifier_map[prop] = identifier_machine.obtain(f"{cls.name}.{prop.name}")

    # endregion

    root.extend(
        _generate_for_constraints_and_patterns(
            identifiers_for_constraints=identifiers_for_constraints
        )
    )

    if len(symbol_table.enumerations) > 0:
        root.append(ET.Comment("Enumerations start here."))

    for enumeration in symbol_table.enumerations:
        root.extend(
            _generate_for_enum(
                enumeration=enumeration,
                identifier_map=identifier_map,
                identifier_machine=identifier_machine,
                name_prefix=name_prefix,
            )
        )

    if len(symbol_table.enumerations) > 0:
        root.append(ET.Comment("Enumerations end here."))

    if len(invariants) > 0:
        root.append(ET.Comment("Constraints start here."))

    for invariant in invariants:
        root.extend(
            _generate_for_invariant(
                invariant=invariant,
                identifier_map=identifier_map,
                identifiers_for_constraints=identifiers_for_constraints,
                identifier_machine=identifier_machine,
            )
        )

    if len(invariants) > 0:
        root.append(ET.Comment("Constraints end here."))

    if len(symbol_table.constrained_primitives) > 0:
        root.append(ET.Comment("Constrained primitives start here."))

    for constrained_primitive in symbol_table.constrained_primitives:
        root.append(
            _generate_for_constrained_primitive(
                constrained_primitive=constrained_primitive,
                identifier_map=identifier_map,
                identifiers_for_constraints=identifiers_for_constraints,
                name_prefix=name_prefix,
            )
        )

    if len(symbol_table.constrained_primitives) > 0:
        root.append(ET.Comment("Constrained primitives end here."))

    for cls in symbol_table.classes:
        if cls.interface is not None:
            root.extend(
                _generate_for_interface(
                    interface=cls.interface,
                    identifier_map=identifier_map,
                    identifiers_for_constraints=identifiers_for_constraints,
                    name_prefix=name_prefix,
                )
            )

    for concrete_cls in symbol_table.concrete_classes:
        root.extend(
            _generate_for_concrete_class(
                cls=concrete_cls,
                identifier_map=identifier_map,
                identifiers_for_constraints=identifiers_for_constraints,
                name_prefix=name_prefix,
            )
        )

    writer = io.StringIO()

    # NOTE (mristin):
    # ElementTree removes the namespaces in the root, so we have to add them back
    # manually.
    root.attrib["xmlns"] = namespace_declarations.main
    for alias, url in namespace_declarations.alias_to_url.items():
        root.attrib[f"xmlns:{alias}"] = url

    _render(element=root, writer=writer, namespace_declarations=namespace_declarations)

    return writer.getvalue(), None


def execute(
    context: run.Context,
    stdout: TextIO,
    stderr: TextIO,
) -> int:
    """
    Execute the generation with the given parameters.

    Return the error code, or 0 if no errors.
    """
    code, errors = _generate(
        symbol_table=context.symbol_table, spec_impls=context.spec_impls
    )
    if errors is not None:
        run.write_error_report(
            message=f"Failed to generate the OPC UA node set "
            f"based on {context.model_path}",
            errors=[context.lineno_columner.error_message(error) for error in errors],
            stderr=stderr,
        )
        return 1

    assert code is not None

    # noinspection SpellCheckingInspection
    pth = context.output_dir / "nodeset.xml"
    try:
        pth.write_text(code, encoding="utf-8")
    except Exception as exception:
        run.write_error_report(
            message=f"Failed to write the OPC UA node set to {pth}",
            errors=[str(exception)],
            stderr=stderr,
        )
        return 1

    stdout.write(f"Code generated to: {context.output_dir}\n")

    return 0



"""Provide code generation for unrolling recursive calls and iterations."""
import abc
import io
import textwrap
from typing import Sequence, List

from icontract import DBC, require

from aas_core_codegen import intermediate
from aas_core_codegen.common import assert_never
from aas_core_codegen.cpp.common import INDENT as I


class Node:
    """Represent a node in the unrolling tree."""

    def __init__(self, text: str, children: Sequence["Node"]):
        self.text = text
        self.children = children


def render(node: Node) -> str:
    """
    Render the node recursively.

    >>> render(Node(text='something', children=[]))
    'something'

    >>> render(Node(text='parent', children=[Node(text='child', children=[])]))
    'parent {\\n  child\\n}'

    >>> render(Node(text='parent', children=[
    ...     Node(text='child1', children=[]),
    ...     Node(text='child2', children=[]),
    ... ]))
    'parent {\\n  child1\\n\\n  child2\\n}'
    """
    if len(node.children) == 0:
        return node.text

    writer = io.StringIO()
    writer.write(node.text)
    writer.write(" {")

    for i, child in enumerate(node.children):
        if i == 0:
            writer.write("\n")
        else:
            writer.write("\n\n")

        writer.write(textwrap.indent(render(child), I))

    writer.write("\n}")

    return writer.getvalue()


class AbstractUnroller(DBC):
    """Generate code to unroll recursion into generic types."""

    @abc.abstractmethod
    def _unroll_primitive_type_annotation(
        self,
        unrollee_expr: str,
        type_annotation: intermediate.PrimitiveTypeAnnotation,
        path: List[str],
        list_loop_level: int,
    ) -> List[Node]:
        """Generate code for the given specific ``type_annotation``."""
        raise NotImplementedError()

    @abc.abstractmethod
    def _unroll_our_type_annotation(
        self,
        unrollee_expr: str,
        type_annotation: intermediate.OurTypeAnnotation,
        path: List[str],
        list_loop_level: int,
    ) -> List[Node]:
        """Generate code for the given specific ``type_annotation``."""
        raise NotImplementedError()

    @abc.abstractmethod
    def _unroll_list_type_annotation(
        self,
        unrollee_expr: str,
        type_annotation: intermediate.ListTypeAnnotation,
        path: List[str],
        list_loop_level: int,
    ) -> List[Node]:
        """Generate code for the given specific ``type_annotation``."""
        raise NotImplementedError()

    @abc.abstractmethod
    def _unroll_optional_type_annotation(
        self,
        unrollee_expr: str,
        type_annotation: intermediate.OptionalTypeAnnotation,
        path: List[str],
        list_loop_level: int,
    ) -> List[Node]:
        """Generate code for the given specific ``type_annotation``."""
        raise NotImplementedError()

    @require(lambda list_loop_level: list_loop_level >= 0)
    def unroll(
        self,
        unrollee_expr: str,
        type_annotation: intermediate.TypeAnnotationUnion,
        path: List[str],
        list_loop_level: int,
    ) -> List[Node]:
        """
        Dispatch the given type annotation to unrolling.

        :param unrollee_expr: Expression of the element to be unrolled
        :param type_annotation: Type annotation corresponding to the ``unrollee_expr``
        :param path:
            Path, as code snippets to be joined by "/" to the ``unrollee_expr``
        :param list_loop_level:
            Depth level of the list loops.

            Level 0 indicates the outer loop.
        """
        if isinstance(type_annotation, intermediate.PrimitiveTypeAnnotation):
            return self._unroll_primitive_type_annotation(
                unrollee_expr=unrollee_expr,
                type_annotation=type_annotation,
                path=path,
                list_loop_level=list_loop_level,
            )

        elif isinstance(type_annotation, intermediate.OurTypeAnnotation):
            return self._unroll_our_type_annotation(
                unrollee_expr=unrollee_expr,
                type_annotation=type_annotation,
                path=path,
                list_loop_level=list_loop_level,
            )

        elif isinstance(type_annotation, intermediate.ListTypeAnnotation):
            return self._unroll_list_type_annotation(
                unrollee_expr=unrollee_expr,
                type_annotation=type_annotation,
                path=path,
                list_loop_level=list_loop_level,
            )

        elif isinstance(type_annotation, intermediate.OptionalTypeAnnotation):
            return self._unroll_optional_type_annotation(
                unrollee_expr=unrollee_expr,
                type_annotation=type_annotation,
                path=path,
                list_loop_level=list_loop_level,
            )
        else:
            assert_never(type_annotation)

        raise AssertionError("Should not have gotten here")



"""Generate the test code for the ``XOrDefault`` methods."""

from typing import List, Optional

from icontract import ensure

from aas_core_codegen import intermediate, naming
from aas_core_codegen.common import Stripped, indent_but_first_line
from aas_core_codegen.csharp import common as csharp_common, naming as csharp_naming
from aas_core_codegen.csharp.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
)


# fmt: off
@ensure(
    lambda result: result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(
    namespace: csharp_common.NamespaceIdentifier,
    symbol_table: intermediate.SymbolTable,
) -> str:
    """
    Generate the test code for the ``XOrDefault`` methods.

    The ``namespace`` indicates the fully-qualified name of the base project.
    """
    blocks = [
        Stripped(
            f"""\
private static void CompareOrRerecordValue(
{I}object value,
{I}string expectedPath)
{{
{I}Nodes.JsonNode got = Aas.Tests.CommonJson.ToJson(
{II}value);

{I}if (Aas.Tests.Common.RecordMode)
{I}{{
{II}string? parent = Path.GetDirectoryName(expectedPath);
{II}if (parent != null)
{II}{{
{III}if (!Directory.Exists(parent))
{III}{{
{IIII}Directory.CreateDirectory(parent);
{III}}}
{II}}}

{II}System.IO.File.WriteAllText(
{III}expectedPath, got.ToJsonString());
{I}}}
{I}else
{I}{{
{II}if (!System.IO.File.Exists(expectedPath))
{II}{{
{III}throw new System.IO.FileNotFoundException(
{IIII}$"The file with the recorded value does not exist: {{expectedPath}}; " +
{IIII}"maybe you want to set the environment " +
{IIII}$"variable {{Aas.Tests.Common.RecordModeEnvironmentVariableName}}?");
{II}}}

{II}Nodes.JsonNode expected = Aas.Tests.CommonJson.ReadFromFile(
{III}expectedPath);

{II}Aas.Tests.CommonJson.CheckJsonNodesEqual(
{III}expected, got, out Aas.Reporting.Error? error);

{II}if (error != null)
{II}{{
{III}Assert.Fail(
{IIII}$"The original value from {{expectedPath}} is unequal the obtain value " +
{IIII}"when serialized to JSON: " +
{IIII}$"{{Reporting.GenerateJsonPath(error.PathSegments)}}: " +
{IIII}error.Cause
{III});
{II}}}
{I}}}
}}"""
        )
    ]  # type: List[Stripped]

    for concrete_cls in symbol_table.concrete_classes:
        cls_name_csharp = csharp_naming.class_name(concrete_cls.name)
        cls_name_json = naming.json_model_type(concrete_cls.name)

        x_or_default_methods = []  # type: List[intermediate.MethodUnion]
        for method in concrete_cls.methods:
            if method.name.endswith("_or_default"):
                x_or_default_methods.append(method)

        for method in x_or_default_methods:
            method_name_csharp = csharp_naming.method_name(method.name)

            result_enum = None  # type: Optional[intermediate.Enumeration]
            assert method.returns is not None, (
                f"Expected all X_or_default to return something, "
                f"but got None for {concrete_cls.name}.{method.name}"
            )

            if isinstance(
                method.returns, intermediate.OurTypeAnnotation
            ) and isinstance(method.returns.our_type, intermediate.Enumeration):
                result_enum = method.returns.our_type

            if result_enum is None:
                value_assignment_snippet = Stripped(
                    f"var value = instance.{method_name_csharp}();"
                )
            else:
                value_assignment_snippet = Stripped(
                    f"""\
string value = Aas.Stringification.ToString(
{I}instance.{method_name_csharp}())
{II}?? throw new System.InvalidOperationException(
{III}"Failed to stringify the enum");"""
                )

            # noinspection SpellCheckingInspection
            blocks.append(
                Stripped(
                    f"""\
[Test]
public void Test_{cls_name_csharp}_{method_name_csharp}_non_default()
{{
{I}Aas.{cls_name_csharp} instance = (
{II}Aas.Tests.CommonJsonization.LoadMaximal{cls_name_csharp}());

{I}{indent_but_first_line(value_assignment_snippet, I)}

{I}CompareOrRerecordValue(
{II}value,
{II}Path.Combine(
{III}Aas.Tests.Common.TestDataDir,
{III}"XOrDefault",
{III}{csharp_common.string_literal(cls_name_json)},
{III}"{method_name_csharp}.non-default.json"));
}}  // public void Test_{cls_name_csharp}_{method_name_csharp}_non_default

[Test]
public void Test_{cls_name_csharp}_{method_name_csharp}_default()
{{
{I}Aas.{cls_name_csharp} instance = (
{II}Aas.Tests.CommonJsonization.LoadMinimal{cls_name_csharp}());

{I}{indent_but_first_line(value_assignment_snippet, I)}

{I}CompareOrRerecordValue(
{II}value,
{II}Path.Combine(
{III}Aas.Tests.Common.TestDataDir,
{III}"XOrDefault",
{III}{csharp_common.string_literal(cls_name_json)},
{III}"{method_name_csharp}.default.json"));
}}  // public void Test_{cls_name_csharp}_{method_name_csharp}_default"""
                )
            )

    blocks_joined = "\n\n".join(blocks)

    return f"""\
{csharp_common.WARNING}

using Aas = {namespace};  // renamed

using Directory = System.IO.Directory;
using Nodes = System.Text.Json.Nodes;
using Path = System.IO.Path;

using NUnit.Framework;  // can't alias

namespace {namespace}.Tests
{{
{I}public class TestXOrDefault
{I}{{
{II}{indent_but_first_line(blocks_joined, II)}
{I}}}  // class TestXOrDefault
}}  // namespace {namespace}.Tests

{csharp_common.WARNING}
"""


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Provide data structures and types for representing regular expressions."""
import abc
import enum
from typing import List, Union, Optional, TypeVar, Generic

from icontract import require, DBC

from aas_core_codegen.parse.tree import FormattedValue

T = TypeVar("T")


class Node(DBC):
    """Represent a generic node in the abstract syntax tree of a regular expression."""

    @abc.abstractmethod
    def accept(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        raise NotImplementedError()

    @abc.abstractmethod
    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        raise NotImplementedError()


class Regex(Node):
    """Represent the root of the AST of a regex."""

    #: List of concatenations joined by "|"
    union: "UnionExpr"

    def __init__(
        self,
        union: "UnionExpr",
    ) -> None:
        """Initialize with the given values."""
        self.union = union

    def accept(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_regex(self)

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_regex(self)


class UnionExpr(Node):
    """
    Represent a regex union of concatenations.

    For example, ``a|b``.
    """

    def __init__(self, uniates: List["Concatenation"]) -> None:
        """Initialize with the given values."""
        self.uniates = uniates

    def accept(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_union_expr(self)

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_union_expr(self)


class Concatenation(Node):
    """Represent a concatenation of one or more regex terms."""

    def __init__(self, concatenants: List["Term"]) -> None:
        """Initialize with the given values."""
        self.concatenants = concatenants

    def accept(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_concatenation(self)

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_concatenation(self)


class SymbolKind(enum.Enum):
    """Represent a symbol in the pattern such as start and end."""

    START = "^"
    END = "$"
    DOT = "."


class Symbol(Node):
    """Represent a special symbol."""

    def __init__(self, kind: SymbolKind) -> None:
        """Initialize with the given values."""
        self.kind = kind

    def accept(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_symbol(self)

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_symbol(self)


TermValueUnion = Union["Group", "Char", "CharSet", FormattedValue, "Symbol"]


class Term(Node):
    """Represent a term in a regex concatenation."""

    # fmt: off
    @require(
        lambda value, quantifier:
        not (
                value is Symbol
                and (
                        value.kind is SymbolKind.START
                        or value.kind is SymbolKind.END
                )
        )
        or quantifier is None,
        "No quantifier for ``^`` and ``$`` symbols"
    )
    # fmt: on
    def __init__(
        self, value: TermValueUnion, quantifier: Optional["Quantifier"]
    ) -> None:
        """Initialize with the given values."""
        self.value = value
        self.quantifier = quantifier

    def accept(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_term(self)

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_term(self)


class Group(Node):
    """Represent a regex group."""

    def __init__(self, union: UnionExpr) -> None:
        """Initialize with the given values."""
        self.union = union

    def accept(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_group(self)

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_group(self)


class Char(Node):
    """Represent a character as-is."""

    #: Decoded character
    character: str

    #: If set, the character should be represented encoded (as ``\xHH``, ``\\uXXXX`` or
    #: ``\UXXXXXXXX``, depending on the code of the :attr:`~character`)
    explicitly_encoded: bool

    @require(lambda character: len(character) == 1)
    def __init__(self, character: str, explicitly_encoded: bool = False) -> None:
        """Initialize with the given values."""
        self.character = character
        self.explicitly_encoded = explicitly_encoded

    def accept(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_char(self)

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_char(self)


class Quantifier(Node):
    """Represent a quantifier applied on a regex term."""

    # fmt: off
    @require(
        lambda minimum, maximum:
        not (maximum is not None)
        or (minimum <= maximum)
    )
    @require(lambda maximum: not (maximum is not None) or maximum >= 0)
    @require(lambda minimum: minimum >= 0)
    # fmt: on
    def __init__(
        self, non_greedy: bool, minimum: int, maximum: Optional[int] = None
    ) -> None:
        """Initialize with the given values."""
        self.non_greedy = non_greedy
        self.minimum = minimum
        self.maximum = maximum

    def accept(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_quantifier(self)

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_quantifier(self)


class CharSet(Node):
    """Represent a regex character set (also sometimes called *character class*)."""

    def __init__(self, complementing: bool, ranges: List["Range"]) -> None:
        """Initialize with the given values."""
        self.complementing = complementing
        self.ranges = ranges

    def accept(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_char_set(self)

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_char_set(self)


class Range(Node):
    """Represent a range in a character set."""

    def __init__(self, start: Char, end: Optional[Char]) -> None:
        """Initialize with the given values."""
        self.start = start
        self.end = end

    def accept(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_range(self)

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_range(self)


class Visitor(DBC):
    """Visit a node and do something about it."""

    def visit(self, node: Node) -> None:
        """Dispatch to the appropriate visiting method."""
        node.accept(self)

    @abc.abstractmethod
    def visit_regex(self, node: Regex) -> None:
        """Visit the ``regex``."""
        raise NotImplementedError()

    @abc.abstractmethod
    def visit_union_expr(self, node: UnionExpr) -> None:
        """Visit the ``union_expr``."""
        raise NotImplementedError()

    @abc.abstractmethod
    def visit_concatenation(self, node: Concatenation) -> None:
        """Visit the ``concatenation``."""
        raise NotImplementedError()

    @abc.abstractmethod
    def visit_symbol(self, node: Symbol) -> None:
        """Visit the ``symbol``."""
        raise NotImplementedError()

    @abc.abstractmethod
    def visit_term(self, node: Term) -> None:
        """Visit the ``term``."""
        raise NotImplementedError()

    @abc.abstractmethod
    def visit_group(self, node: Group) -> None:
        """Visit the ``group``."""
        raise NotImplementedError()

    @abc.abstractmethod
    def visit_char(self, node: Char) -> None:
        """Visit the ``char``."""
        raise NotImplementedError()

    @abc.abstractmethod
    def visit_quantifier(self, node: Quantifier) -> None:
        """Visit the ``quantifier``."""
        raise NotImplementedError()

    @abc.abstractmethod
    def visit_char_set(self, node: CharSet) -> None:
        """Visit the ``char_set``."""
        raise NotImplementedError()

    @abc.abstractmethod
    def visit_range(self, node: Range) -> None:
        """Visit the ``range``."""
        raise NotImplementedError()


class Transformer(Generic[T], DBC):
    """Transform the regular expression into something else."""

    def transform(self, node: Node) -> T:
        """Dispatch the ``node`` to the appropriate transforming method."""
        return node.transform(self)

    @abc.abstractmethod
    def transform_regex(self, node: Regex) -> T:
        """Transform the ``regex`` into something."""
        raise NotImplementedError()

    @abc.abstractmethod
    def transform_union_expr(self, node: UnionExpr) -> T:
        """Transform the ``union_expr`` into something."""
        raise NotImplementedError()

    @abc.abstractmethod
    def transform_concatenation(self, node: Concatenation) -> T:
        """Transform the ``concatenation`` into something."""
        raise NotImplementedError()

    @abc.abstractmethod
    def transform_symbol(self, node: Symbol) -> T:
        """Transform the ``symbol`` into something."""
        raise NotImplementedError()

    @abc.abstractmethod
    def transform_term(self, node: Term) -> T:
        """Transform the ``term`` into something."""
        raise NotImplementedError()

    @abc.abstractmethod
    def transform_group(self, node: Group) -> T:
        """Transform the ``group`` into something."""
        raise NotImplementedError()

    @abc.abstractmethod
    def transform_char(self, node: Char) -> T:
        """Transform the ``char`` into something."""
        raise NotImplementedError()

    @abc.abstractmethod
    def transform_quantifier(self, node: Quantifier) -> T:
        """Transform the ``quantifier`` into something."""
        raise NotImplementedError()

    @abc.abstractmethod
    def transform_char_set(self, node: CharSet) -> T:
        """Transform the ``char_set`` into something."""
        raise NotImplementedError()

    @abc.abstractmethod
    def transform_range(self, node: Range) -> T:
        """Transform the ``range`` into something."""
        raise NotImplementedError()



"""Generate C++ code for common functions."""

from aas_core_codegen.cpp.aas_common import _generate

generate_header = _generate.generate_header
generate_implementation = _generate.generate_implementation



"""Generate the invariant verifiers from the intermediate representation."""
import io
import textwrap
from typing import (
    Tuple,
    Optional,
    List,
    Sequence,
    Mapping,
    Union,
)

from icontract import ensure, require

from aas_core_codegen import intermediate, specific_implementations
from aas_core_codegen.common import (
    Error,
    Stripped,
    assert_never,
    Identifier,
    indent_but_first_line,
    wrap_text_into_lines,
    assert_union_without_excluded,
)
from aas_core_codegen.python.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
)
from aas_core_codegen.intermediate import type_inference as intermediate_type_inference
from aas_core_codegen.parse import tree as parse_tree, retree as parse_retree
from aas_core_codegen.python import (
    common as python_common,
    naming as python_naming,
    description as python_description,
    transpilation as python_transpilation,
)


# region Verify


def verify(
    spec_impls: specific_implementations.SpecificImplementations,
    verification_functions: Sequence[intermediate.Verification],
) -> Optional[List[str]]:
    """Verify all the implementation snippets related to verification."""
    errors = []  # type: List[str]

    expected_keys = []  # type: List[specific_implementations.ImplementationKey]

    for func in verification_functions:
        if isinstance(func, intermediate.ImplementationSpecificVerification):
            expected_keys.append(
                specific_implementations.ImplementationKey(
                    f"Verification/{func.name}.py"
                ),
            )

    for key in expected_keys:
        if key not in spec_impls:
            errors.append(f"The implementation snippet is missing for: {key}")

    if len(errors) == 0:
        return None

    return errors


# endregion

# region Generate


class _PatternVerificationTranspiler(
    parse_tree.RestrictedTransformer[Tuple[Optional[Stripped], Optional[Error]]]
):
    """Transpile a statement of a pattern verification into Python."""

    @ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
    def _transform_joined_str_values(
        self, values: Sequence[Union[str, parse_tree.FormattedValue]]
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        """Transform the values of a joined string to a Python string literal."""
        # If we do not need interpolation, simply return the string literals
        # joined together by newlines.
        needs_interpolation = any(
            isinstance(value, parse_tree.FormattedValue) for value in values
        )
        if not needs_interpolation:
            return (
                Stripped(
                    python_common.string_literal(
                        "".join(value for value in values)  # type: ignore
                    )
                ),
                None,
            )

        parts = []  # type: List[str]

        # NOTE (mristin, 2022-09-30):
        # See which quotes occur more often in the non-interpolated parts, so that we
        # pick the escaping scheme which will result in as little escapes as possible.
        double_quotes_count = 0
        single_quotes_count = 0

        for value in values:
            if isinstance(value, str):
                double_quotes_count += value.count('"')
                single_quotes_count += value.count("'")

            elif isinstance(value, parse_tree.FormattedValue):
                pass
            else:
                assert_never(value)

        # Pick the escaping scheme
        if single_quotes_count <= double_quotes_count:
            enclosing = "'"
            quoting = python_common.StringQuoting.SINGLE_QUOTES
        else:
            enclosing = '"'
            quoting = python_common.StringQuoting.DOUBLE_QUOTES

        for value in values:
            if isinstance(value, str):
                parts.append(
                    python_common.string_literal(
                        value,
                        quoting=quoting,
                        without_enclosing=True,
                        duplicate_curly_brackets=True,
                    )
                )

            elif isinstance(value, parse_tree.FormattedValue):
                code, error = self.transform(value.value)
                if error is not None:
                    return None, error

                assert code is not None

                assert (
                    "\n" not in code
                ), f"New-lines are not expected in formatted values, but got: {code}"

                parts.append(f"{{{code}}}")
            else:
                assert_never(value)

        writer = io.StringIO()
        writer.write("f")
        writer.write(enclosing)
        for part in parts:
            writer.write(part)
        writer.write(enclosing)

        return Stripped(writer.getvalue()), None

    @ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
    def transform_constant(
        self, node: parse_tree.Constant
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        if isinstance(node.value, str):
            # NOTE (mristin, 2022-06-11):
            # We assume that all the string constants are valid regular expressions.

            regex, parse_error = parse_retree.parse(values=[node.value])
            if parse_error is not None:
                regex_line, pointer_line = parse_retree.render_pointer(
                    parse_error.cursor
                )

                return (
                    None,
                    Error(
                        node.original_node,
                        f"The string constant could not be parsed "
                        f"as a regular expression: \n"
                        f"{parse_error.message}\n"
                        f"{regex_line}\n"
                        f"{pointer_line}",
                    ),
                )

            assert regex is not None

            # NOTE (mristin, 2022-09-30):
            # Strictly speaking, this is a joined string with a single value, a string
            # literal.
            return self._transform_joined_str_values(
                values=parse_retree.render(regex=regex)
            )
        else:
            raise AssertionError(f"Unexpected {node=}")

    @ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
    def transform_name(
        self, node: parse_tree.Name
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        return Stripped(python_naming.variable_name(node.identifier)), None

    @ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
    def transform_joined_str(
        self, node: parse_tree.JoinedStr
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        regex, parse_error = parse_retree.parse(values=node.values)
        if parse_error is not None:
            regex_line, pointer_line = parse_retree.render_pointer(parse_error.cursor)

            return (
                None,
                Error(
                    node.original_node,
                    f"The joined string could not be parsed "
                    f"as a regular expression: \n"
                    f"{parse_error.message}\n"
                    f"{regex_line}\n"
                    f"{pointer_line}",
                ),
            )

        assert regex is not None

        return self._transform_joined_str_values(
            values=parse_retree.render(regex=regex)
        )

    def transform_assignment(
        self, node: parse_tree.Assignment
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        assert isinstance(node.target, parse_tree.Name)
        variable = python_naming.variable_name(node.target.identifier)
        code, error = self.transform(node.value)
        if error is not None:
            return None, error
        assert code is not None

        return Stripped(f"{variable} = {code}"), None


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _transpile_pattern_verification(
    verification: intermediate.PatternVerification,
    qualified_module_name: python_common.QualifiedModuleName,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the verification function that checks the regular expressions."""
    # NOTE (mristin, 2022-09-30):
    # We assume that we performed all the checks at the intermediate stage.

    construct_name = python_naming.function_name(
        Identifier(f"_construct_{verification.name}")
    )

    blocks = []  # type: List[Stripped]

    # region Construct block

    writer = io.StringIO()
    writer.write(
        f"""\
# noinspection SpellCheckingInspection
def {construct_name}() -> Pattern[str]:
"""
    )

    transpiler = _PatternVerificationTranspiler()

    for i, stmt in enumerate(verification.parsed.body):
        if i == len(verification.parsed.body) - 1:
            break

        code, error = transpiler.transform(stmt)
        if error is not None:
            return None, error
        assert code is not None

        writer.write(textwrap.indent(code, I))
        writer.write("\n")

    if len(verification.parsed.body) >= 2:
        writer.write("\n")

    pattern_expr, error = transpiler.transform(verification.pattern_expr)
    if error is not None:
        return None, error
    assert pattern_expr is not None

    # A pragmatic heuristics for breaking lines
    if len(pattern_expr) < 50:
        writer.write(textwrap.indent(f"return re.compile({pattern_expr})", I))
    else:
        writer.write(
            textwrap.indent(
                f"""\
return re.compile(
{I}{indent_but_first_line(pattern_expr, I)}
)""",
                I,
            )
        )

    blocks.append(Stripped(writer.getvalue()))

    # endregion

    # region Initialize the regex

    regex_name = python_naming.constant_name(Identifier(f"_regex_{verification.name}"))

    blocks.append(Stripped(f"{regex_name} = {construct_name}()"))

    assert len(verification.arguments) == 1
    assert isinstance(
        verification.arguments[0].type_annotation, intermediate.PrimitiveTypeAnnotation
    )
    # noinspection PyUnresolvedReferences
    assert (
        verification.arguments[0].type_annotation.a_type
        == intermediate.PrimitiveType.STR
    )

    arg_name = python_naming.argument_name(verification.arguments[0].name)

    function_name = python_naming.function_name(verification.name)

    writer = io.StringIO()
    writer.write(
        f"""\
def {function_name}({arg_name}: str) -> bool:
"""
    )

    if verification.description is not None:
        (
            docstring,
            docstring_errors,
        ) = python_description.generate_docstring_for_signature(
            description=verification.description,
            context=python_description.Context(
                qualified_module_name=qualified_module_name,
                module=Identifier("verification"),
                cls_or_enum=None,
            ),
        )
        if docstring_errors is not None:
            return None, Error(
                verification.description.parsed.node,
                "Failed to generate the docstring",
                docstring_errors,
            )

        assert docstring is not None

        writer.write(textwrap.indent(docstring, I))
        writer.write("\n")

    writer.write(f"{I}return {regex_name}.match({arg_name}) is not None")

    blocks.append(Stripped(writer.getvalue()))

    # endregion

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n\n")

        writer.write(block)

    return Stripped(writer.getvalue()), None


class _TranspilableVerificationTranspiler(python_transpilation.Transpiler):
    """Transpile the body of a :class:`.TranspilableVerification`."""

    # fmt: off
    @require(
        lambda environment, verification:
        all(
            environment.find(arg.name) is not None
            for arg in verification.arguments
        ),
        "All arguments defined in the environment"
    )
    # fmt: on
    def __init__(
        self,
        type_map: Mapping[
            parse_tree.Node, intermediate_type_inference.TypeAnnotationUnion
        ],
        environment: intermediate_type_inference.Environment,
        symbol_table: intermediate.SymbolTable,
        verification: intermediate.TranspilableVerification,
    ) -> None:
        """Initialize with the given values."""
        python_transpilation.Transpiler.__init__(
            self, type_map=type_map, environment=environment
        )

        self._symbol_table = symbol_table

        self._argument_name_set = frozenset(arg.name for arg in verification.arguments)

    def transform_name(
        self, node: parse_tree.Name
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        if node.identifier in self._variable_name_set:
            return Stripped(python_naming.variable_name(node.identifier)), None

        if node.identifier in self._argument_name_set:
            return Stripped(python_naming.argument_name(node.identifier)), None

        if node.identifier in self._symbol_table.constants_by_name:
            constant_name = python_naming.constant_name(node.identifier)
            return Stripped(f"aas_constants.{constant_name}"), None

        if node.identifier in self._symbol_table.verification_functions_by_name:
            return Stripped(python_naming.function_name(node.identifier)), None

        our_type = self._symbol_table.find_our_type(name=node.identifier)
        if isinstance(our_type, intermediate.Enumeration):
            return (
                Stripped(f"aas_types.{python_naming.enum_name(node.identifier)}"),
                None,
            )

        return None, Error(
            node.original_node,
            f"We can not determine how to transpile the name {node.identifier!r} "
            f"to Python. We could not find it neither in the constants, nor in "
            f"verification functions, nor as an enumeration. "
            f"If you expect this name to be transpilable, please contact "
            f"the developers.",
        )


def _transpile_transpilable_verification(
    verification: intermediate.TranspilableVerification,
    symbol_table: intermediate.SymbolTable,
    environment: intermediate_type_inference.Environment,
    qualified_module_name: python_common.QualifiedModuleName,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Transpile a verification function."""
    # fmt: off
    type_inference, error = (
        intermediate_type_inference.infer_for_verification(
            verification=verification,
            base_environment=environment
        )
    )
    # fmt: on

    if error is not None:
        return None, error

    assert type_inference is not None

    transpiler = _TranspilableVerificationTranspiler(
        type_map=type_inference.type_map,
        environment=type_inference.environment_with_args,
        symbol_table=symbol_table,
        verification=verification,
    )

    body = []  # type: List[Stripped]
    for node in verification.parsed.body:
        stmt, error = transpiler.transform(node)
        if error is not None:
            return None, Error(
                verification.parsed.node,
                f"Failed to transpile the verification function {verification.name!r}",
                [error],
            )

        assert stmt is not None
        body.append(stmt)

    writer = io.StringIO()

    function_name = python_naming.function_name(verification.name)

    if verification.returns is None:
        return_type = "None"
    else:
        return_type = python_common.generate_type(
            type_annotation=verification.returns, types_module=Identifier("aas_types")
        )

    arg_defs = []  # type: List[Stripped]
    for arg in verification.arguments:
        arg_type = python_common.generate_type(
            arg.type_annotation, types_module=Identifier("aas_types")
        )
        arg_name = python_naming.argument_name(arg.name)
        arg_defs.append(Stripped(f"{arg_name}: {arg_type}"))

    if len(arg_defs) == 0:
        writer.write(
            f"""\
def {function_name}() -> {return_type}:"""
        )
    else:
        writer.write(
            f"""\
def {function_name}(
"""
        )

        for i, arg_def in enumerate(arg_defs):
            if i > 0:
                writer.write(",\n")
            writer.write(textwrap.indent(arg_def, I))

        writer.write("\n")
        writer.write(
            f"""\
) -> {return_type}:"""
        )

    docstring = None  # type: Optional[Stripped]
    if verification.description is not None:
        (
            docstring,
            docstring_errors,
        ) = python_description.generate_docstring_for_signature(
            description=verification.description,
            context=python_description.Context(
                qualified_module_name=qualified_module_name,
                module=Identifier("verification"),
                cls_or_enum=None,
            ),
        )
        if docstring_errors is not None:
            return None, Error(
                verification.description.parsed.node,
                "Failed to generate the docstring",
                docstring_errors,
            )

        assert docstring is not None

        writer.write("\n")
        writer.write(textwrap.indent(docstring, I))

    writer.write(f"\n{I}# pylint: disable=all")

    if docstring is None and len(body) == 0:
        writer.write(f"\n{I}pass")

    for stmt in body:
        writer.write("\n")
        writer.write(textwrap.indent(stmt, I))

    return Stripped(writer.getvalue()), None


class _InvariantTranspiler(python_transpilation.Transpiler):
    def __init__(
        self,
        type_map: Mapping[
            parse_tree.Node, intermediate_type_inference.TypeAnnotationUnion
        ],
        environment: intermediate_type_inference.Environment,
        symbol_table: intermediate.SymbolTable,
    ) -> None:
        """Initialize with the given values."""
        python_transpilation.Transpiler.__init__(
            self, type_map=type_map, environment=environment
        )

        self._symbol_table = symbol_table

    def transform_name(
        self, node: parse_tree.Name
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        if node.identifier in self._variable_name_set:
            return Stripped(python_naming.variable_name(node.identifier)), None

        if node.identifier == "self":
            # The ``that`` refers to the argument of the verification function.
            return Stripped("that"), None

        if node.identifier in self._symbol_table.constants_by_name:
            constant_name = python_naming.constant_name(node.identifier)
            return Stripped(f"aas_constants.{constant_name}"), None

        if node.identifier in self._symbol_table.verification_functions_by_name:
            return Stripped(python_naming.function_name(node.identifier)), None

        our_type = self._symbol_table.find_our_type(name=node.identifier)
        if isinstance(our_type, intermediate.Enumeration):
            return (
                Stripped(f"aas_types.{python_naming.enum_name(node.identifier)}"),
                None,
            )

        return None, Error(
            node.original_node,
            f"We can not determine how to transpile the name {node.identifier!r} "
            f"to Python. We could not find it neither in the local variables, "
            f"nor in the global constants, nor in verification functions, "
            f"nor as an enumeration. If you expect this name to be transpilable, "
            f"please contact the developers.",
        )


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _transpile_invariant(
    invariant: intermediate.Invariant,
    symbol_table: intermediate.SymbolTable,
    environment: intermediate_type_inference.Environment,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Translate the invariant from the meta-model into Python code."""
    # fmt: off
    type_map, inference_error = (
        intermediate_type_inference.infer_for_invariant(
            invariant=invariant,
            environment=environment
        )
    )
    # fmt: on

    if inference_error is not None:
        return None, inference_error

    assert type_map is not None

    transpiler = _InvariantTranspiler(
        type_map=type_map,
        environment=environment,
        symbol_table=symbol_table,
    )

    expr, error = transpiler.transform(invariant.parsed.body)
    if error is not None:
        return None, error

    assert expr is not None

    writer = io.StringIO()
    if len(expr) > 50 or "\n" in expr:
        writer.write("if not (\n")
        writer.write(textwrap.indent(expr, I))
        writer.write("\n):\n")
    else:
        no_parenthesis_type_in_this_context = (
            parse_tree.Index,
            parse_tree.Name,
            parse_tree.Member,
            parse_tree.MethodCall,
            parse_tree.FunctionCall,
        )

        if isinstance(invariant.parsed.body, no_parenthesis_type_in_this_context):
            not_expr = f"not {expr}"
        else:
            not_expr = f"not ({expr})"

        writer.write(f"if {not_expr}:\n")

    writer.write(f"{I}yield Error(\n")

    # NOTE (mristin, 2022-09-30):
    # We need to wrap the description in multiple literals as a single long
    # string literal is often too much for the readability.
    invariant_description_lines = wrap_text_into_lines(invariant.description)

    for i, literal in enumerate(invariant_description_lines):
        if i < len(invariant_description_lines) - 1:
            writer.write(f"{II}{python_common.string_literal(literal)} +\n")
        else:
            writer.write(f"{II}{python_common.string_literal(literal)}\n")
            writer.write(f"{I})")

    return Stripped(writer.getvalue()), None


OurTypeExceptEnumeration = Union[
    intermediate.ConstrainedPrimitive,
    intermediate.AbstractClass,
    intermediate.ConcreteClass,
]
assert_union_without_excluded(
    original_union=intermediate.OurType,
    subset_union=OurTypeExceptEnumeration,
    excluded=[intermediate.Enumeration],
)


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate_verify_property_snippet(
    prop: intermediate.Property,
    generator_for_loop_variables: python_common.GeneratorForLoopVariables,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """
    Generate the snippet to transform a property to verification errors.

    Return an empty string if there is nothing to be verified for the given property.
    """
    # NOTE (mristin, 2022-10-01):
    # Instead of writing here a complex but general solution with unrolling we choose
    # to provide a simple, but limited, solution. First, the meta-model is quite
    # limited itself at the moment, so the complexity of the general solution is not
    # warranted. Second, we hope that there will be fewer bugs in the simple solution
    # which is particularly important at this early adoption stage.
    #
    # We anticipate that in the future we will indeed need a general and complex
    # solution. Here are just some thoughts on how to approach it:
    # * Leave the pattern matching to produce more readable code for simple cases,
    # * Unroll only in case of composite types and optional composite types.

    type_anno = (
        prop.type_annotation
        if not isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation)
        else prop.type_annotation.value
    )

    if isinstance(type_anno, intermediate.OptionalTypeAnnotation):
        return None, Error(
            prop.parsed.node,
            "We currently implemented verification based on a very limited "
            "pattern matching due to code simplicity. We did not handle "
            "the case of nested optional values. Please contact "
            "the developers if you need this functionality.",
        )
    elif isinstance(type_anno, intermediate.ListTypeAnnotation):
        if isinstance(type_anno.items, intermediate.OptionalTypeAnnotation):
            return None, Error(
                prop.parsed.node,
                "We currently implemented verification based on a very limited "
                "pattern matching due to code simplicity. We did not handle "
                "the case of lists of optional values. Please contact "
                "the developers if you need this functionality.",
            )
        elif isinstance(type_anno.items, intermediate.ListTypeAnnotation):
            return None, Error(
                prop.parsed.node,
                "We currently implemented verification based on a very limited "
                "pattern matching due to code simplicity. We did not handle "
                "the case of lists of lists. Please contact "
                "the developers if you need this functionality.",
            )
        else:
            pass
    else:
        pass

    stmts = []  # type: List[Stripped]

    prop_name = python_naming.property_name(prop.name)
    prop_name_literal = python_common.string_literal(prop_name)

    if isinstance(type_anno, intermediate.PrimitiveTypeAnnotation):
        # There is nothing that we check for primitive types explicitly. The values
        # of the primitive properties are checked at the level of class invariants.
        return Stripped(""), None
    elif isinstance(type_anno, intermediate.OurTypeAnnotation):
        if isinstance(type_anno.our_type, intermediate.Enumeration):
            # We rely on mypy to check for valid enumerations, so we do not check
            # the enumerations on our side.
            return Stripped(""), None

        elif isinstance(type_anno.our_type, intermediate.ConstrainedPrimitive):
            function_name = python_naming.function_name(
                Identifier(f"verify_{type_anno.our_type.name}")
            )

            for_error_in_verify = f"for error in {function_name}(that.{prop_name})"
            # Heuristic to break the lines, very rudimentary
            if len(for_error_in_verify) > 70:
                for_error_in_verify = f"""\
for error in {function_name}(
{II}that.{prop_name}
)"""

            stmts.append(
                Stripped(
                    f"""\
{for_error_in_verify}:
{I}error.path._prepend(
{II}PropertySegment(
{III}that,
{III}{prop_name_literal}
{II})
{I})
{I}yield error"""
                )
            )

        elif isinstance(
            type_anno.our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
        ):
            for_error_in_self_transform = (
                f"for error in self.transform(that.{prop_name})"
            )
            # Heuristic to break the lines, very rudimentary
            if len(for_error_in_self_transform) > 70:
                for_error_in_self_transform = f"""\
for error in self.transform(
{II}that.{prop_name}
)"""

            stmts.append(
                Stripped(
                    f"""\
{for_error_in_self_transform}:
{I}error.path._prepend(
{II}PropertySegment(
{III}that,
{III}{prop_name_literal}
{II})
{I})
{I}yield error"""
                )
            )
        else:
            assert_never(type_anno.our_type)

    elif isinstance(type_anno, intermediate.ListTypeAnnotation):
        assert not isinstance(
            type_anno.items,
            (intermediate.OptionalTypeAnnotation, intermediate.ListTypeAnnotation),
        ), (
            "We chose to implement only a very limited pattern matching; "
            "see the note above in the code."
        )

        # NOTE (mristin, 2022-10-01):
        # We only descend into our classes here.
        if not isinstance(type_anno.items, intermediate.OurTypeAnnotation):
            return Stripped(""), None

        loop_variable = next(generator_for_loop_variables)

        for_i_item_in_that_prop = (
            f"for i, {loop_variable} in enumerate(that.{prop_name})"
        )

        # Rudimentary heuristics for line breaking
        if len(for_i_item_in_that_prop) > 70:
            for_i_item_in_that_prop = f"""\
for i, {loop_variable} in enumerate(
{II}that.{prop_name}
)"""

        for_error_in_self_transform = f"for error in self.transform({loop_variable})"
        if len(for_error_in_self_transform) > 70:
            for_error_in_self_transform = f"""\
for error in self.transform(
{II}{loop_variable}
)"""

        stmts.append(
            Stripped(
                f"""\
{for_i_item_in_that_prop}:
{I}{indent_but_first_line(for_error_in_self_transform, I)}:
{II}error.path._prepend(
{III}IndexSegment(
{IIII}that.{prop_name},
{IIII}i
{III})
{II})
{II}error.path._prepend(
{III}PropertySegment(
{IIII}that,
{IIII}{prop_name_literal}
{III})
{II})
{II}yield error"""
            )
        )

    else:
        assert_never(type_anno)

    verify_block = Stripped("\n".join(stmts))

    if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
        return (
            Stripped(
                f"""\
if that.{prop_name} is not None:
{I}{indent_but_first_line(verify_block, I)}"""
            ),
            None,
        )

    return verify_block, None


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate_transform_for_class(
    cls: intermediate.ConcreteClass,
    symbol_table: intermediate.SymbolTable,
    base_environment: intermediate_type_inference.Environment,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the transform method to errors for the given concrete class."""
    errors = []  # type: List[Error]
    blocks = []  # type: List[Stripped]

    environment = intermediate_type_inference.MutableEnvironment(
        parent=base_environment
    )

    assert environment.find(Identifier("self")) is None
    environment.set(
        identifier=Identifier("self"),
        type_annotation=intermediate_type_inference.OurTypeAnnotation(our_type=cls),
    )

    for invariant in cls.invariants:
        invariant_code, error = _transpile_invariant(
            invariant=invariant, symbol_table=symbol_table, environment=environment
        )
        if error is not None:
            errors.append(
                Error(
                    cls.parsed.node,
                    f"Failed to transpile the invariant of the class {cls.name!r}",
                    [error],
                )
            )
            continue

        assert invariant_code is not None

        blocks.append(invariant_code)

    if len(errors) > 0:
        return None, errors

    # NOTE (mristin, 2022-10-14):
    # We need to generate unique loop variable for each loop since Python tracks
    # the variables in function scope, not block scope.
    generator_for_loop_variables = python_common.GeneratorForLoopVariables()

    for prop in cls.properties:
        block, error = _generate_verify_property_snippet(
            prop=prop, generator_for_loop_variables=generator_for_loop_variables
        )
        if error is not None:
            errors.append(error)
        else:
            assert block is not None
            if block != "":
                blocks.append(block)

    if len(errors) > 0:
        return None, errors

    cls_name = python_naming.class_name(cls.name)

    if len(blocks) == 0:
        blocks.append(
            Stripped(
                f"""\
# No verification has been defined for {cls_name}.
return
# For this uncommon return-yield construction, see:
# https://stackoverflow.com/questions/13243766/how-to-define-an-empty-generator-function
# noinspection PyUnreachableCode
yield"""
            )
        )

    transform_name = python_naming.method_name(Identifier(f"transform_{cls.name}"))

    writer = io.StringIO()
    writer.write(
        f"""\
# noinspection PyMethodMayBeStatic
def {transform_name}(
{II}self,
{II}that: aas_types.{cls_name}
) -> Iterator[Error]:
"""
    )

    for i, stmt in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")
        writer.write(textwrap.indent(stmt, I))

    return Stripped(writer.getvalue()), None


def _generate_transformer(
    symbol_table: intermediate.SymbolTable,
    base_environment: intermediate_type_inference.Environment,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate a transformer to double-dispatch an instance to errors."""
    errors = []  # type: List[Error]

    blocks = []  # type: List[Stripped]

    # The abstract classes are directly dispatched by the transformer,
    # so we do not need to handle them separately.

    for cls in symbol_table.concrete_classes:
        if cls.is_implementation_specific:
            transform_key = specific_implementations.ImplementationKey(
                f"Verification/transform_{cls.name}.py"
            )

            implementation = spec_impls.get(transform_key, None)
            if implementation is None:
                errors.append(
                    Error(
                        cls.parsed.node,
                        f"The transformation snippet is missing "
                        f"for the implementation-specific "
                        f"class {cls.name}: {transform_key}",
                    )
                )
                continue

            blocks.append(spec_impls[transform_key])
        else:
            block, cls_errors = _generate_transform_for_class(
                cls=cls,
                symbol_table=symbol_table,
                base_environment=base_environment,
            )
            if cls_errors is not None:
                errors.extend(cls_errors)
            else:
                assert block is not None
                blocks.append(block)

    if len(errors) > 0:
        return None, errors

    writer = io.StringIO()
    writer.write(
        f"""\
class _Transformer(
{II}aas_types.AbstractTransformer[
{III}Iterator[Error]
{II}]
):
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")
        writer.write(textwrap.indent(block, I))

    return Stripped(writer.getvalue()), None


def _generate_verify_constrained_primitive(
    constrained_primitive: intermediate.ConstrainedPrimitive,
    symbol_table: intermediate.SymbolTable,
    base_environment: intermediate_type_inference.Environment,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the verify function for the constrained primitives."""
    errors = []  # type: List[Error]
    blocks = []  # type: List[Stripped]

    environment = intermediate_type_inference.MutableEnvironment(
        parent=base_environment
    )

    assert environment.find(Identifier("self")) is None
    environment.set(
        identifier=Identifier("self"),
        type_annotation=intermediate_type_inference.OurTypeAnnotation(
            our_type=constrained_primitive
        ),
    )

    for invariant in constrained_primitive.invariants:
        invariant_code, error = _transpile_invariant(
            invariant=invariant, symbol_table=symbol_table, environment=environment
        )
        if error is not None:
            errors.append(
                Error(
                    constrained_primitive.parsed.node,
                    f"Failed to transpile the invariant of "
                    f"the constrained primitive {constrained_primitive.name!r}",
                    [error],
                )
            )
            continue

        assert invariant_code is not None

        blocks.append(invariant_code)

    if len(errors) > 0:
        return None, errors

    no_verification_specified = False
    if len(blocks) == 0:
        no_verification_specified = True
        blocks.append(
            Stripped(
                """\
# There is no verification specified.
return

# Empty generator according to:
# https://stackoverflow.com/a/13243870/1600678
# noinspection PyUnreachableCode
yield"""
            )
        )

    function_name = python_naming.function_name(
        Identifier(f"verify_{constrained_primitive.name}")
    )

    that_type = python_common.PRIMITIVE_TYPE_MAP[constrained_primitive.constrainee]

    writer = io.StringIO()

    if no_verification_specified:
        # NOTE (mristin, 2022-10-02):
        # We provide a function for evolvability even though it does nothing.
        writer.write("# noinspection PyUnusedLocal\n")

    writer.write(
        f"""\
def {function_name}(
{II}that: {that_type}
) -> Iterator[Error]:
{I}\"\"\"Verify the constraints of :paramref:`that`.\"\"\"
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")
        writer.write(textwrap.indent(block, I))

    assert len(errors) == 0
    return Stripped(writer.getvalue()), None


def _generate_module_docstring(
    symbol_table: intermediate.SymbolTable,
    qualified_module_name: python_common.QualifiedModuleName,
) -> Stripped:
    """Generate the docstring for the module."""
    docstring_blocks = [
        Stripped("Verify that the instances of the meta-model satisfy the invariants.")
    ]  # type: List[Stripped]

    first_cls = (
        symbol_table.concrete_classes[0]
        if len(symbol_table.concrete_classes) > 0
        else None
    )  # type: Optional[intermediate.ConcreteClass]

    if first_cls is not None:
        cls_name = python_naming.class_name(first_cls.name)
        an_instance_variable = python_naming.variable_name(Identifier("an_instance"))

        docstring_blocks.append(
            Stripped(
                f"""\
Here is an example how to verify an instance of :py:class:`{qualified_module_name}.types.{cls_name}`:

.. code-block::

    import {qualified_module_name}.types as aas_types
    import {qualified_module_name}.verification as aas_verification

    {an_instance_variable} = aas_types.{cls_name}(
        # ... some constructor arguments ...
    )

    for error in aas_verification.verify({an_instance_variable}):
        print(f"{{error.cause}} at: {{error.path}}")"""
            )
        )

    # endregion

    if len(docstring_blocks) == 1:
        doc_escaped = docstring_blocks[0].replace('"""', '\\"\\"\\"')
        docstring = f'"""{doc_escaped}"""'
    else:
        doc_escaped = ("\n\n".join(docstring_blocks)).replace('"""', '\\"\\"\\"')
        docstring = f'''\
"""
{doc_escaped}
"""'''

    return Stripped(docstring)


# fmt: off
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
@ensure(
    lambda result:
    not (result[0] is not None) or result[0].endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(
    symbol_table: intermediate.SymbolTable,
    qualified_module_name: python_common.QualifiedModuleName,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """
    Generate the Python code for verification based on the symbol table.

    The ``qualified_module_name`` indicates the fully-qualified name of the base module.
    """
    # region Module docstring
    blocks = [
        _generate_module_docstring(
            symbol_table=symbol_table, qualified_module_name=qualified_module_name
        ),
        python_common.WARNING,
        Stripped(
            f"""\
import math
import re
import struct
import sys
from typing import (
{I}Callable,
{I}Iterable,
{I}Iterator,
{I}List,
{I}Mapping,
{I}Optional,
{I}Pattern,
{I}Sequence,
{I}Set,
{I}Union
)

if sys.version_info >= (3, 8):
{I}from typing import Final
else:
{I}from typing_extensions import Final

from {qualified_module_name} import (
{I}constants as aas_constants,
{I}types as aas_types,
)"""
        ),
        Stripped(
            f"""\
class PropertySegment:
{I}\"\"\"Represent a property access on a path to an erroneous value.\"\"\"

{I}#: Instance containing the property
{I}instance: Final[aas_types.Class]

{I}#: Name of the property
{I}name: Final[str]

{I}def __init__(
{III}self,
{III}instance: aas_types.Class,
{III}name: str
{I}) -> None:
{II}\"\"\"Initialize with the given values.\"\"\"
{II}self.instance = instance
{II}self.name = name

{I}def __str__(self) -> str:
{II}return f'.{{self.name}}'"""
        ),
        Stripped(
            f"""\
class IndexSegment:
{I}\"\"\"Represent an index access on a path to an erroneous value.\"\"\"

{I}#: Sequence containing the item at :py:attr:`~index`
{I}sequence: Final[Sequence[aas_types.Class]]

{I}#: Index of the item
{I}index: Final[int]

{I}def __init__(
{III}self,
{III}sequence: Sequence[aas_types.Class],
{III}index: int
{I}) -> None:
{II}\"\"\"Initialize with the given values.\"\"\"
{II}self.sequence = sequence
{II}self.index = index

{I}def __str__(self) -> str:
{II}return f'[{{self.index}}]'"""
        ),
        Stripped("Segment = Union[PropertySegment, IndexSegment]"),
        Stripped(
            f"""\
class Path:
{I}\"\"\"Represent the relative path to the erroneous value.\"\"\"

{I}def __init__(self) -> None:
{II}\"\"\"Initialize as an empty path.\"\"\"
{II}self._segments = []  # type: List[Segment]

{I}@property
{I}def segments(self) -> Sequence[Segment]:
{II}\"\"\"Get the segments of the path.\"\"\"
{II}return self._segments

{I}def _prepend(self, segment: Segment) -> None:
{II}\"\"\"Insert the :paramref:`segment` in front of other segments.\"\"\"
{II}self._segments.insert(0, segment)

{I}def __str__(self) -> str:
{II}return "".join(str(segment) for segment in self._segments)"""
        ),
        Stripped(
            f"""\
class Error:
{I}\"\"\"Represent a verification error in the data.\"\"\"

{I}#: Human-readable description of the error
{I}cause: Final[str]

{I}#: Path to the erroneous value
{I}path: Final[Path]

{I}def __init__(self, cause: str) -> None:
{II}\"\"\"Initialize as an error with an empty path.\"\"\"
{II}self.cause = cause
{II}self.path = Path()

{I}def __repr__(self) -> str:
{II}return f"Error(path={{self.path}}, cause={{self.cause}})\""""
        ),
    ]  # type: List[Stripped]

    errors = []  # type: List[Error]

    base_environment = intermediate_type_inference.populate_base_environment(
        symbol_table=symbol_table
    )

    for verification in symbol_table.verification_functions:
        if isinstance(verification, intermediate.ImplementationSpecificVerification):
            implementation_key = specific_implementations.ImplementationKey(
                f"Verification/{verification.name}.py"
            )

            implementation = spec_impls.get(implementation_key, None)
            if implementation is None:
                errors.append(
                    Error(
                        None,
                        f"The snippet for the verification function "
                        f"{verification.name!r} is missing: {implementation_key}",
                    )
                )
            else:
                blocks.append(implementation)

        elif isinstance(verification, intermediate.PatternVerification):
            implementation, error = _transpile_pattern_verification(
                verification=verification, qualified_module_name=qualified_module_name
            )

            if error is not None:
                errors.append(error)
            else:
                assert implementation is not None
                blocks.append(implementation)

        elif isinstance(verification, intermediate.TranspilableVerification):
            implementation, error = _transpile_transpilable_verification(
                verification=verification,
                symbol_table=symbol_table,
                environment=base_environment,
                qualified_module_name=qualified_module_name,
            )

            if error is not None:
                errors.append(error)
            else:
                assert implementation is not None
                blocks.append(implementation)

        else:
            assert_never(verification)

    transformer_block, transformer_errors = _generate_transformer(
        symbol_table=symbol_table,
        base_environment=base_environment,
        spec_impls=spec_impls,
    )
    if transformer_errors is not None:
        errors.extend(transformer_errors)
    else:
        assert transformer_block is not None
        blocks.append(transformer_block)

    blocks.append(Stripped("_TRANSFORMER = _Transformer()"))

    blocks.append(
        Stripped(
            f"""\
def verify(
{II}that: aas_types.Class
) -> Iterator[Error]:
{I}\"\"\"
{I}Verify the constraints of :paramref:`that` recursively.

{I}:param that: instance whose constraints we want to verify
{I}:yield: constraint violations
{I}\"\"\"
{I}yield from _TRANSFORMER.transform(that)"""
        )
    )

    for our_type in symbol_table.our_types:
        if isinstance(our_type, intermediate.Enumeration):
            # NOTE (mristin, 2022-10-01):
            # We do not verify the enumerations explicitly in Python as mypy
            # is capable enough to spot invalid enum literals.
            pass

        elif isinstance(our_type, intermediate.ConstrainedPrimitive):
            (
                constrained_primitive_block,
                constrained_primitive_errors,
            ) = _generate_verify_constrained_primitive(
                constrained_primitive=our_type,
                symbol_table=symbol_table,
                base_environment=base_environment,
            )

            if constrained_primitive_errors is not None:
                errors.extend(constrained_primitive_errors)
            else:
                assert constrained_primitive_block is not None
                blocks.append(constrained_primitive_block)

        elif isinstance(
            our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
        ):
            # NOTE (mristin, 2022-10-01):
            # We provide a general dispatch function for the most abstract
            # class ``Class``.
            pass
        else:
            assert_never(our_type)

    blocks.append(python_common.WARNING)

    if len(errors) > 0:
        return None, errors

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue(), None


# endregion
